# SPDX-License-Identifier: Apache-2.0
import logging
import tarfile

import salt.exceptions
import saltext.vmware.utils.cluster as utils_cluster
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.datacenter as utils_datacenter
import saltext.vmware.utils.esxi as utils_esxi

# pylint: disable=no-name-in-module
try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

log = logging.getLogger(__name__)


def power_cycle_vm(virtual_machine, action="on"):
    """
    Powers on/off a virtual machine specified by its name.

    virtual_machine
        vim.VirtualMachine object to power on/off virtual machine

    action
        Operation option to power on/off the machine
    """
    if action == "powered-on":
        try:
            task = virtual_machine.PowerOnVM_Task()
            task_name = "power on"
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)
    elif action == "powered-off":
        try:
            task = virtual_machine.PowerOffVM_Task()
            task_name = "power off"
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)
    elif action == "suspend":
        try:
            task = virtual_machine.SuspendVM_Task()
            task_name = "suspend"
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)
    elif action == "reset":
        try:
            task = virtual_machine.ResetVM_Task()
            task_name = "reset"
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)
    else:
        raise salt.exceptions.ArgumentValueError("The given action is not supported")
    try:
        utils_common.wait_for_task(
            task, utils_common.get_managed_object_name(virtual_machine), task_name
        )
    except salt.exceptions.VMwareFileNotFoundError as exc:
        raise salt.exceptions.VMwarePowerOnError(
            " ".join(
                [
                    "An error occurred during power",
                    "operation, a file was not found: {}".format(exc),
                ]
            )
        )
    return virtual_machine


def create_vm(vm_name, vm_config_spec, folder_object, resourcepool_object, host_object=None):
    """
    Creates virtual machine from config spec

    vm_name
        Virtual machine name to be created

    vm_config_spec
        Virtual Machine Config Spec object

    folder_object
        vm Folder managed object reference

    resourcepool_object
        Resource pool object where the machine will be created

    host_object
        Host object where the machine will be placed (optional)

    return
        Virtual Machine managed object reference
    """
    try:
        if host_object and isinstance(host_object, vim.HostSystem):
            task = folder_object.CreateVM_Task(
                vm_config_spec, pool=resourcepool_object, host=host_object
            )
        else:
            task = folder_object.CreateVM_Task(vm_config_spec, pool=resourcepool_object)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    vm_object = utils_common.wait_for_task(task, vm_name, "CreateVM Task", 10, "info")
    return vm_object


def clone_vm(vm_name, folder_object, template, clone_config_spec):
    """
    Creates virtual machine from config spec

    Returns Virtual Machine managed object reference

    vm_name
        Virtual machine name to be created

    folder_object
        Virtual Machine Folder managed object reference

    template
        Vitual Machine template to clone from

    clone_config_spec
        Clone Config Spec object
    """
    try:
        task = template.CloneVM_Task(folder_object, vm_name, clone_config_spec)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    vm_object = utils_common.wait_for_task(task, vm_name, "CloneVM Task", 10, "info")
    return vm_object


def register_vm(datacenter, name, vmx_path, resourcepool_object, host_object=None):
    """
    Registers a virtual machine to the inventory with the given vmx file, on success
    it returns the vim.VirtualMachine managed object reference

    datacenter
        Datacenter object of the virtual machine, vim.Datacenter object

    name
        Name of the virtual machine

    vmx_path:
        Full path to the vmx file, datastore name should be included

    resourcepool
        Placement resource pool of the virtual machine, vim.ResourcePool object

    host
        Placement host of the virtual machine, vim.HostSystem object
    """
    try:
        if host_object:
            task = datacenter.vmFolder.RegisterVM_Task(
                path=vmx_path,
                name=name,
                asTemplate=False,
                host=host_object,
                pool=resourcepool_object,
            )
        else:
            task = datacenter.vmFolder.RegisterVM_Task(
                path=vmx_path, name=name, asTemplate=False, pool=resourcepool_object
            )
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    try:
        vm_ref = utils_common.wait_for_task(task, name, "RegisterVM Task")
    except salt.exceptions.VMwareFileNotFoundError as exc:
        raise salt.exceptions.VMwareVmRegisterError(
            "An error occurred during registration operation, the "
            "configuration file was not found: {}".format(exc)
        )
    return vm_ref


def update_vm(vm_ref, vm_config_spec):
    """
    Updates the virtual machine configuration with the given object

    vm_ref
        Virtual machine managed object reference

    vm_config_spec
        Virtual machine config spec object to update
    """
    vm_name = utils_common.get_managed_object_name(vm_ref)
    log.trace("Updating vm '%s'", vm_name)
    try:
        task = vm_ref.ReconfigVM_Task(vm_config_spec)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    vm_ref = utils_common.wait_for_task(task, vm_name, "ReconfigureVM Task")
    return vm_ref


def delete_vm(vm_ref):
    """
    Destroys the virtual machine

    vm_ref
        Managed object reference of a virtual machine object
    """
    vm_name = utils_common.get_managed_object_name(vm_ref)
    log.trace("Destroying vm '%s'", vm_name)
    try:
        task = vm_ref.Destroy_Task()
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    utils_common.wait_for_task(task, vm_name, "Destroy Task")


def unregister_vm(vm_ref):
    """
    Destroys the virtual machine

    vm_ref
        Managed object reference of a virtual machine object
    """
    vm_name = utils_common.get_managed_object_name(vm_ref)
    log.trace("Destroying vm '%s'", vm_name)
    try:
        vm_ref.UnregisterVM()
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        raise salt.exceptions.VMwareRuntimeError(exc.msg)


def list_vms(
    service_instance,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
):
    """
    Returns a list of VMs associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain VMs.

    datacenter_name
        The datacenter name. Default is None.

    cluster_name
        The cluster name - used to restrict the VMs retrieved. Only used if
        the datacenter is set.  This argument is optional.

    host_name
        The host_name to be retrieved. Default is None.
    """
    properties = ["name", "config"]

    if cluster_name and not datacenter_name:
        raise salt.exceptions.ArgumentValueError(
            "Must specify the datacenter when specifying the cluster"
        )

    if not datacenter_name:
        # Assume the root folder is the starting point
        start_point = utils_common.get_root_folder(service_instance)
    else:
        start_point = utils_datacenter.get_datacenter(service_instance, datacenter_name)
        if cluster_name:
            # Retrieval to test if cluster exists. Cluster existence only makes
            # sense if the datacenter has been specified
            properties.append("parent")

    if host_name:
        host = utils_common.get_mor_by_property(service_instance, vim.HostSystem, host_name)
        properties.append("runtime.host")
        log.trace("Retrieved host: %s", host)
    else:
        host = None

    # Search for the objects
    vms = utils_common.get_mors_with_properties(
        service_instance,
        vim.VirtualMachine,
        container_ref=start_point,
        property_list=properties,
    )

    items = []
    for vm in vms:
        if vm.get("config") and not vm["config"].template:
            if host_name:
                if host == vm["runtime.host"]:
                    items.append(vm["name"])
            else:
                items.append(vm["name"])
    return items


def list_vm_templates(service_instance):
    """
    Returns a list of template VMs associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain VMs.
    """
    items = []
    vms = utils_common.get_mors_with_properties(service_instance, vim.VirtualMachine)
    for vm in vms:
        if vm["config"].template:
            items.append(vm["name"])
    return items


def get_vm_by_property(
    service_instance,
    name,
    datacenter=None,
    vm_properties=None,
    traversal_spec=None,
    parent_ref=None,
):
    """
    Get virtual machine properties based on the traversal specs and properties list,
    returns Virtual Machine object with properties.

    service_instance
        Service instance object to access vCenter

    name
        Name of the virtual machine.

    datacenter
        Datacenter name

    vm_properties
        List of vm properties.

    traversal_spec
        Traversal Spec object(s) for searching.

    parent_ref
        Container Reference object for searching under a given object.
    """
    if datacenter and not parent_ref:
        parent_ref = utils_datacenter.get_datacenter(service_instance, datacenter)
    if not vm_properties:
        vm_properties = [
            "name",
            "config.hardware.device",
            "summary.storage.committed",
            "summary.storage.uncommitted",
            "summary.storage.unshared",
            "layoutEx.file",
            "config.guestFullName",
            "config.guestId",
            "guest.net",
            "config.hardware.memoryMB",
            "config.hardware.numCPU",
            "config.files.vmPathName",
            "summary.runtime.powerState",
            "guest.toolsStatus",
        ]
    vm_list = utils_common.get_mors_with_properties(
        service_instance,
        vim.VirtualMachine,
        vm_properties,
        container_ref=parent_ref,
        traversal_spec=traversal_spec,
    )
    vm_formatted = [vm for vm in vm_list if vm["name"] == name]
    if not vm_formatted:
        raise salt.exceptions.VMwareObjectRetrievalError("The virtual machine was not found.")
    elif len(vm_formatted) > 1:
        raise salt.exceptions.VMwareMultipleObjectsError(
            " ".join(
                [
                    "Multiple virtual machines were found with the"
                    "same name, please specify a container."
                ]
            )
        )
    return vm_formatted[0]


def get_placement(service_instance, datacenter, placement=None):
    """
    To create a virtual machine a resource pool needs to be supplied, we would like to use the strictest as possible.

    datacenter
        Name of the datacenter

    placement
        Dictionary with the placement info, cluster, host resource pool name

    return
        Resource pool, cluster and host object if any applies
    """
    log.trace("Retrieving placement information")
    resourcepool_object, placement_object = None, None
    if "host" in placement:
        host_objects = utils_esxi.get_hosts(
            service_instance, datacenter_name=datacenter, host_names=[placement["host"]]
        )
        if not host_objects:
            raise salt.exceptions.VMwareObjectRetrievalError(
                " ".join(
                    [
                        "The specified host",
                        "{} cannot be found.".format(placement["host"]),
                    ]
                )
            )
        try:
            host_props = utils_common.get_properties_of_managed_object(
                host_objects[0], properties=["resourcePool"]
            )
            resourcepool_object = host_props["resourcePool"]
        except vmodl.query.InvalidProperty:
            traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
                path="parent",
                skip=True,
                type=vim.HostSystem,
                selectSet=[
                    vmodl.query.PropertyCollector.TraversalSpec(
                        path="resourcePool", skip=False, type=vim.ClusterComputeResource
                    )
                ],
            )
            resourcepools = utils_common.get_mors_with_properties(
                service_instance,
                vim.ResourcePool,
                container_ref=host_objects[0],
                property_list=["name"],
                traversal_spec=traversal_spec,
            )
            if resourcepools:
                resourcepool_object = resourcepools[0]["object"]
            else:
                raise salt.exceptions.VMwareObjectRetrievalError(
                    "The resource pool of host {} cannot be found.".format(placement["host"])
                )
        placement_object = host_objects[0]
    elif "resourcepool" in placement:
        resourcepool_objects = utils_common.get_resource_pools(
            service_instance, [placement["resourcepool"]], datacenter_name=datacenter
        )
        if len(resourcepool_objects) > 1:
            raise salt.exceptions.VMwareMultipleObjectsError(
                " ".join(
                    [
                        "Multiple instances are available of the",
                        "specified host {}.".format(placement["host"]),
                    ]
                )
            )
        resourcepool_object = resourcepool_objects[0]
        res_props = utils_common.get_properties_of_managed_object(
            resourcepool_object, properties=["parent"]
        )
        if "parent" in res_props:
            placement_object = res_props["parent"]
        else:
            raise salt.exceptions.VMwareObjectRetrievalError(
                " ".join(["The resource pool's parent", "object is not defined"])
            )
    elif "cluster" in placement:
        datacenter_object = utils_datacenter.get_datacenter(service_instance, datacenter)
        cluster_object = utils_cluster.get_cluster(datacenter_object, placement["cluster"])
        clus_props = utils_common.get_properties_of_managed_object(
            cluster_object, properties=["resourcePool"]
        )
        if "resourcePool" in clus_props:
            resourcepool_object = clus_props["resourcePool"]
        else:
            raise salt.exceptions.VMwareObjectRetrievalError(
                " ".join(["The cluster's resource pool", "object is not defined"])
            )
        placement_object = cluster_object
    else:
        # We are checking the schema for this object, this exception should never be raised
        raise salt.exceptions.VMwareObjectRetrievalError(" ".join(["Placement is not defined."]))
    return (resourcepool_object, placement_object)


def get_folder(service_instance, datacenter, placement, base_vm_name=None):
    """
    Returns a Folder Object

    service_instance
        Service instance object

    datacenter
        Name of the datacenter

    placement
        Placement dictionary

    base_vm_name
        Existing virtual machine name (for cloning)
    """
    log.trace("Retrieving folder information")
    if base_vm_name:
        vm_object = get_vm_by_property(service_instance, base_vm_name, vm_properties=["name"])
        vm_props = utils_common.get_properties_of_managed_object(vm_object, properties=["parent"])
        if "parent" in vm_props:
            folder_object = vm_props["parent"]
        else:
            raise salt.exceptions.VMwareObjectRetrievalError(
                " ".join(["The virtual machine parent", "object is not defined"])
            )
    # elif "folder" in placement:
    #     folder_objects = get_folders(
    #         service_instance, [placement["folder"]], datacenter
    #     )
    #     if len(folder_objects) > 1:
    #         raise salt.exceptions.VMwareMultipleObjectsError(
    #             " ".join(
    #                 [
    #                     "Multiple instances are available of the",
    #                     "specified folder {}".format(placement["folder"]),
    #                 ]
    #             )
    #         )
    #     folder_object = folder_objects[0]
    elif datacenter:
        datacenter_object = utils_datacenter.get_datacenter(service_instance, datacenter)
        dc_props = utils_common.get_properties_of_managed_object(
            datacenter_object, properties=["vmFolder"]
        )
        if "vmFolder" in dc_props:
            folder_object = dc_props["vmFolder"]
        else:
            raise salt.exceptions.VMwareObjectRetrievalError(
                "The datacenter vm folder object is not defined"
            )
    return folder_object


def read_ovf_file(ovf_path):
    """
    Read in OVF file.

    ovf_path
        Path to ovf file
    """
    try:
        with open(ovf_path) as ovf_file:
            return ovf_file.read()
    except Exception:
        exit(f"Could not read file: {ovf_path}")


def read_ovf_from_ova(ova_path):
    """
    Read in OVF file from OVA.

    ova_path
        Path to ova file
    """
    try:
        with tarfile.open(ova_path) as tf:
            for entry in tf:
                if entry.name.endswith(".ovf"):
                    ovf = tf.extractfile(entry)
                    return ovf.read().decode()
    except Exception:
        exit(f"Could not read file: {ova_path}")


def get_disk_size(vm):
    """
    Returns total disk size in bytes from the virtual machine object.

    vm
        Virtual Machine Object from which to obtain mac address.
    """
    size = 0.0
    if vm.config:
        for device in vm.config.hardware.device:
            if isinstance(device, vim.vm.device.VirtualDisk):
                size += device.capacityInBytes

    return size


def get_network(vm):
    """
    Returns network from a virtual machine object.

    vm
        Virtual Machine Object from which to obtain mac address.
    """
    network = {}
    for device in vm.guest.net:
        network[device.macAddress] = {}
        network[device.macAddress]["ipv4"] = []
        network[device.macAddress]["ipv6"] = []
        for address in device.ipAddress:
            if "::" in address:
                network[device.macAddress]["ipv6"].append(address)
            else:
                network[device.macAddress]["ipv4"].append(address)
    return network


def get_mac_address(vm):
    """
    Returns mac addresses from a virtual machine object.

    vm
        Virtual Machine Object from which to obtain mac address.
    """
    mac_address = []
    if vm.config:
        for device in vm.config.hardware.device:
            if isinstance(device, vim.vm.device.VirtualEthernetCard):
                mac_address.append(device.macAddress)
    return mac_address


def options_order_list(vm, order):
    """
    Converts a text order into internal representation for setting the boot order.

    vm
        Reference to virtual machine to check options on.

    order
        (List of strings) Boot order of devices. Acceptable strings: cdrom, disk, ethernet, floppy
    """

    boot_order_list = []
    for device_name in order:
        if device_name == "cdrom":
            cdrom = [
                device
                for device in vm.config.hardware.device
                if isinstance(device, vim.vm.device.VirtualCdrom)
            ]
            if cdrom:
                boot_order_list.append(vim.vm.BootOptions.BootableCdromDevice())
        elif device_name == "disk":
            hdd = [
                device
                for device in vm.config.hardware.device
                if isinstance(device, vim.vm.device.VirtualDisk)
            ]
            if hdd:
                boot_order_list.append(vim.vm.BootOptions.BootableDiskDevice(deviceKey=hdd[0].key))
        elif device_name == "ethernet":
            ether = [
                device
                for device in vm.config.hardware.device
                if isinstance(device, vim.vm.device.VirtualEthernetCard)
            ]
            if ether:
                boot_order_list.append(
                    vim.vm.BootOptions.BootableEthernetDevice(deviceKey=ether[0].key)
                )
        elif device_name == "floppy":
            floppy = [
                device
                for device in vm.config.hardware.device
                if isinstance(device, vim.vm.device.VirtualFloppy)
            ]
            if floppy:
                boot_order_list.append(vim.vm.BootOptions.BootableFloppyDevice())
        else:
            raise salt.exceptions.VMwareRuntimeError("invalid order name")
    return boot_order_list


def compare_boot_order_list(new, current):
    """
    Compares current boot order list and input boot order list.

    new
        List of vim bootable device objects.

    current
        List of vim bootable device objects.
    """
    if len(new) == len(current):
        for provided, existing in zip(new, current):
            if provided.deviceKey != existing.deviceKey:
                return False
        return True
    else:
        return False


def compare_boot_options(input_opts, current):
    """
    Compares current boot options and input options.

    input_opts
        (dict) Dictionary of virtual machine boot options.

    current
        Pyvmomi virtual machine boot options object.
    """

    lists_same = compare_boot_order_list(input_opts["bootOrder"], current.bootOrder)
    if (
        lists_same
        and current.bootDelay == input_opts["bootDelay"]
        and current.enterBIOSSetup == input_opts["enterBIOSSetup"]
        and current.bootRetryEnabled == input_opts["bootRetryEnabled"]
        and current.bootRetryDelay == input_opts["bootRetryDelay"]
        and current.efiSecureBootEnabled == input_opts["efiSecureBootEnabled"]
    ):
        return True
    else:
        return False


def boot_options_dif(input_opts, current):
    """
    Returns the difference between two dictionaries of virtual machine boot options.

    input_opts
        (dict) Dictionary of virtual machine boot options.

    current
        Pyvmomi virtual machine boot options object.
    """
    ret = {}
    lists_same = compare_boot_order_list(input_opts["bootOrder"], current.bootOrder)
    if not lists_same:
        old = [i.deviceKey for i in current.bootOrder]
        new = [i.deviceKey for i in input_opts["bootOrder"]]
        ret["order"] = {"old": old, "new": new}
    if not current.bootDelay == input_opts["bootDelay"]:
        ret["delay"] = {"old": current.bootDelay, "new": input_opts["bootDelay"]}
    if not current.enterBIOSSetup == input_opts["enterBIOSSetup"]:
        ret["enter_bois_setup"] = {
            "old": current.enterBIOSSetup,
            "new": input_opts["enterBIOSSetup"],
        }
    if not current.bootRetryEnabled == input_opts["bootRetryEnabled"]:
        ret["retry_enabled"] = {
            "old": current.bootRetryEnabled,
            "new": input_opts["bootRetryEnabled"],
        }
    if not current.bootRetryDelay == input_opts["bootRetryDelay"]:
        ret["retry_delay"] = {
            "old": current.bootRetryDelay,
            "new": input_opts["bootRetryDelay"],
        }
    if not current.efiSecureBootEnabled == input_opts["efiSecureBootEnabled"]:
        ret["efi_secure_boot_enabled"] = {
            "old": current.efiSecureBootEnabled,
            "new": input_opts["efiSecureBootEnabled"],
        }
    return ret


def change_boot_options(vm, input_opts):
    """
    Changes boot options on given vm.

    vm
        reference to virtual machine to change options on.

    input_opts
        (dict) Dictionary of virtual machine boot options.
    """
    vm_conf = vim.vm.ConfigSpec()
    vm_conf.bootOptions = vim.vm.BootOptions(**input_opts)
    task = vm.ReconfigVM_Task(vm_conf)
    utils_common.wait_for_task(task, vm.name, "configure boot options")

    return {"status": "changed"}


def create_snapshot(vm_ref, snapshot_name, description, memory, quiesce):
    """
    Create virtual machine snapshot.

    vm_ref
        Reference to virtual machine.

    snapshot_name
        The name for the snapshot being created. Not unique

    description
        Description for the snapshot.

    memory
        (boolean) If TRUE, a dump of the internal state of the virtual machine (basically a memory dump) is included in the snapshot.

    quiesce
        (boolean) If TRUE and the virtual machine is powered on when the snapshot is taken, VMware Tools is used to quiesce the file system in the virtual machine.
    """
    task = vm_ref.CreateSnapshot_Task(snapshot_name, description, memory, quiesce)
    ret = utils_common.wait_for_task(task, vm_ref.name, "create snapshot")
    return ret


def snapshot_recursive(snapshot_tree_root, snaps):
    """
    Recursively appends all snapshots in a snapshot tree.

    snapshot_tree_root
        Root node of snapshot tree.

    snaps
        List of snapshot info.
    """
    for ss in snapshot_tree_root:
        current = {
            "creation_time": str(ss.createTime),
            "description": ss.description,
            "id": ss.id,
            "name": ss.name,
            "state": ss.state,
            "quiesced": ss.quiesced,
        }
        snaps.append(current)
        if ss.childSnapshotList:
            snaps = snapshot_recursive(ss.childSnapshotList, snaps)
    return snaps


def snapshot_recursive_search(snapshot_tree_root, snapshot_name, snapshot_id):
    """
    Recursively appends all snapshots in a snapshot tree.

    snapshot_tree_root
        Root node of snapshot tree.

    snapshot_name
        Name of snapshot to find.

    snapshot_id
        id of snapshot to find.
    """
    for ss in snapshot_tree_root:
        if snapshot_id is None:
            if snapshot_name == ss.name:
                return ss
            elif ss.childSnapshotList:
                snaps = snapshot_recursive_search(ss.childSnapshotList, snapshot_name, snapshot_id)
            else:
                return None
        else:
            if snapshot_id and snapshot_id == ss.id and snapshot_name == ss.name:
                return ss
            elif ss.childSnapshotList:
                snaps = snapshot_recursive_search(ss.childSnapshotList, snapshot_name, snapshot_id)
            else:
                return None
    return snaps


def get_snapshots(vm_ref):
    """
    Returns list of snapshot info for a given vm.

    vm_ref
        Reference to a virtual machine.
    """
    snaps = []
    tree = vm_ref.snapshot
    if hasattr(tree, "rootSnapshotList") and len(tree.rootSnapshotList) > 0:
        snaps = snapshot_recursive(tree.rootSnapshotList, snaps)
    else:
        snaps = {"msg": "no snapshots"}
    return snaps


def get_snapshot(vm_ref, snapshot_name, snapshot_id):
    """"""
    tree = vm_ref.snapshot
    if hasattr(tree, "rootSnapshotList") and tree.rootSnapshotList:
        snap = snapshot_recursive_search(tree.rootSnapshotList, snapshot_name, snapshot_id)
    else:
        snap = {"msg": "no snapshots"}
    return snap


def destroy_snapshot(snapshot, remove_children):
    """
    Destroy a given snapshot.

    snapshot
        Reference to a vim.vm.Snapshot object.

    remove_children
        Remove subtree of snapshots.
    """
    vm_name = snapshot.vm.name
    task = snapshot.RemoveSnapshot_Task(remove_children)
    ret = utils_common.wait_for_task(task, vm_name, "remove snapshot")
    return ret


def relocate(vm, host, datastore, pool):
    """
    Relocates a virtual machine to the location specified.

    vm
        Reference to virtual machine.

    host
        Reference to new host.

    datastore
        Reference to datastore

    Pool
        Reference to resource pool.
    """
    relocate_spec = vim.vm.RelocateSpec(host=host, datastore=datastore, pool=pool)
    task = vm.RelocateVM_Task(relocate_spec)
    utils_common.wait_for_task(task, vm.name, "move vm")
    return task.info.state



"""
Salt State file to create/update/delete an IP Address Block

Example:

.. code-block:: yaml

    create_ip_block:
      nsxt_ip_blocks.present:
        - name: Create IP Block
          hostname: {{ pillar['nsxt_manager_hostname'] }}
          username: {{ pillar['nsxt_manager_username'] }}
          password: {{ pillar['nsxt_manager_password'] }}
          cert: {{ pillar['nsxt_manager_certificate'] }}
          verify_ssl: <False/True>
          display_name: <ip block name>
          description: <ip block description>
          tags:
            - tag: <tag-key-1>
              scope: <tag-value-1>
            - tag: <tag-key-2>
              scope: <tag-value-2>
          cidr: <cidr>
"""
import logging

log = logging.getLogger(__name__)

__virtual_name__ = "nsxt_ip_blocks"


def __virtual__():
    """
    Only load if the nsxt_ip_blocks module is available in __salt__
    """
    return (
        __virtual_name__ if "nsxt_ip_blocks.get" in __salt__ else False,
        "'nsxt_ip_blocks' binary not found on system",
    )


def _create_state_response(name, old_state, new_state, result, comment):
    state_response = dict()
    state_response["name"] = name
    state_response["result"] = result
    state_response["comment"] = comment
    state_response["changes"] = dict()
    if new_state or old_state:
        state_response["changes"]["old"] = old_state
        state_response["changes"]["new"] = new_state

    return state_response


def _check_for_updates(existing_ip_block, input_dict):
    updatable_keys = ["cidr", "description", "tags"]

    is_updatable = False

    # check if any updatable field has different value from the existing one
    for key in updatable_keys:
        existing_val = existing_ip_block.get(key)
        input_val = input_dict.get(key)
        if existing_val != input_val and input_val:
            return True

    return is_updatable


def _fill_input_dict_with_existing_info(existing_ip_block, input_dict):
    for key in dict(existing_ip_block).keys():
        if key not in input_dict:
            input_dict[key] = existing_ip_block[key]


def present(
    name,
    display_name,
    cidr,
    hostname,
    username,
    password,
    cert=None,
    verify_ssl=True,
    cert_common_name=None,
    description=None,
    tags=None,
):
    """
    Creates/Updates(if present with the same name) an IP Address Block

    .. code-block:: yaml

        create_ip_block:
          nsxt_ip_blocks.present:
            - name: Create IP Block
              hostname: {{ pillar['nsxt_manager_hostname'] }}
              username: {{ pillar['nsxt_manager_username'] }}
              password: {{ pillar['nsxt_manager_password'] }}
              cert: {{ pillar['nsxt_manager_certificate'] }}
              verify_ssl: <False/True>
              display_name: <ip block name>
              description: <ip block description>
              tags:
                - tag: <tag-key-1>
                  scope: <tag-value-1>
                - tag: <tag-key-2>
                  scope: <tag-value-2>
              cidr: <cidr>

    name
        The Operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    display_name
        The name using which IP Address Block will be created. In case of update the provided display_name should be
        unique else the method will raise an error

    description
        (Optional) description for the IP Address Block

    tags
        (Optional) Opaque identifiers meaningful to the API user. Maximum 30 tags can be associated:

        .. code-block:: yaml

            tags:
                - tag: <tag-key-1>
                  scope: <tag-value-1>
                - tag: <tag-key-2>
                  scope: <tag-value-2>

    cidr
        Represents network address and the prefix length which will be associated with a layer-2 broadcast domain
    """

    input_dict = {
        "display_name": display_name,
        "cidr": cidr,
        "description": description,
        "tags": tags,
    }

    log.info("Checking if IP Block with name %s is present", display_name)
    get_ip_blocks_response = __salt__["nsxt_ip_blocks.get_by_display_name"](
        hostname,
        username,
        password,
        display_name,
        cert=cert,
        verify_ssl=verify_ssl,
        cert_common_name=cert_common_name,
    )

    if "error" in get_ip_blocks_response:
        return _create_state_response(name, None, None, False, get_ip_blocks_response["error"])

    ip_blocks = get_ip_blocks_response["results"]

    if len(ip_blocks) > 1:
        log.info("Multiple instances found for the provided display name %s", display_name)
        return _create_state_response(
            name,
            None,
            None,
            False,
            "Multiple IP Blocks found for the provided display name {}".format(display_name),
        )

    existing_ip_block = ip_blocks[0] if len(ip_blocks) > 0 else None

    if __opts__.get("test"):
        log.info("present is called with test option")
        if existing_ip_block:
            return _create_state_response(
                name,
                None,
                None,
                None,
                "State present will update IP Block with name {}".format(display_name),
            )
        else:
            return _create_state_response(
                name,
                None,
                None,
                None,
                "State present will create IP Block with name {}".format(display_name),
            )
    if existing_ip_block:
        is_update_required = _check_for_updates(existing_ip_block, input_dict)

        if is_update_required:
            _fill_input_dict_with_existing_info(existing_ip_block, input_dict)

            log.info("IP Block found with name %s", display_name)
            updated_ip_block = __salt__["nsxt_ip_blocks.update"](
                ip_block_id=existing_ip_block["id"],
                revision=existing_ip_block["_revision"],
                hostname=hostname,
                username=username,
                password=password,
                cert=cert,
                verify_ssl=verify_ssl,
                cert_common_name=cert_common_name,
                display_name=input_dict.get("display_name"),
                cidr=input_dict.get("cidr"),
                description=input_dict.get("description"),
                tags=input_dict.get("tags"),
            )

            if "error" in updated_ip_block:
                return _create_state_response(name, None, None, False, updated_ip_block["error"])

            return _create_state_response(
                name,
                existing_ip_block,
                updated_ip_block,
                True,
                "Updated IP Block {}".format(display_name),
            )
        else:
            log.info("All fields are same as existing IP Address Block %s", display_name)
            return _create_state_response(
                name, None, None, True, "IP Address Block exists already, no action to perform"
            )
    else:
        log.info("No IP Block found with name %s", display_name)
        created_ip_block = __salt__["nsxt_ip_blocks.create"](
            cidr=input_dict.get("cidr"),
            hostname=hostname,
            username=username,
            password=password,
            display_name=input_dict.get("display_name"),
            cert=cert,
            verify_ssl=verify_ssl,
            cert_common_name=cert_common_name,
            description=input_dict.get("description"),
            tags=input_dict.get("tags"),
        )

        if "error" in created_ip_block:
            return _create_state_response(name, None, None, False, created_ip_block["error"])

        return _create_state_response(
            name, None, created_ip_block, True, "Created IP Block {}".format(display_name)
        )


def absent(
    name,
    display_name,
    hostname,
    username,
    password,
    cert=None,
    verify_ssl=True,
    cert_common_name=None,
):
    """
    Deletes an IP Address Block of provided name (if present)

    .. code-block:: yaml

        delete_ip_block:
          nsxt_ip_blocks.absent:
            - name: Delete IP Block
              hostname: <hostname>
              username: <username>
              password: <password>
              cert: <certificate>
              verify_ssl: <False/True>
              display_name: <ip block name>

    name
        The Operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    display_name
        Display name of IP Address Block to delete
    """

    log.info("Checking if IP Address Block with name %s is present", display_name)

    get_ip_blocks_response = __salt__["nsxt_ip_blocks.get_by_display_name"](
        hostname,
        username,
        password,
        display_name,
        cert=cert,
        verify_ssl=verify_ssl,
        cert_common_name=cert_common_name,
    )

    if "error" in get_ip_blocks_response:
        return _create_state_response(name, None, None, False, get_ip_blocks_response["error"])

    ip_blocks = get_ip_blocks_response["results"]

    if len(ip_blocks) > 1:
        log.info("Multiple instances found for the provided display name %s", display_name)
        return _create_state_response(
            name,
            None,
            None,
            False,
            "Multiple IP Blocks found for the provided display name {}".format(display_name),
        )

    existing_ip_block = ip_blocks[0] if len(ip_blocks) > 0 else None

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if existing_ip_block:
            return _create_state_response(
                name,
                None,
                None,
                None,
                "State absent will delete IP Block with name {}".format(display_name),
            )
        else:
            return _create_state_response(
                name,
                None,
                None,
                None,
                "State absent will do nothing as no IP Block found with name {}".format(
                    display_name
                ),
            )

    if existing_ip_block:
        log.info("IP Address Block found with name %s", display_name)
        deleted_response = __salt__["nsxt_ip_blocks.delete"](
            existing_ip_block["id"],
            hostname,
            username,
            password,
            cert=cert,
            verify_ssl=verify_ssl,
            cert_common_name=cert_common_name,
        )

        if "error" in deleted_response:
            return _create_state_response(name, None, None, False, deleted_response["error"])

        return _create_state_response(
            name, existing_ip_block, None, True, "Deleted IP Block {}".format(display_name)
        )
    else:
        log.info("No IP Address Block found with name %s", display_name)
        return _create_state_response(
            name, None, None, True, "No IP Address Block found with name {}".format(display_name)
        )



# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import ssl

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from requests.exceptions import RequestException
from requests.exceptions import SSLError

# pylint: disable=no-name-in-module
try:
    from pyVim import connect
    from pyVim.connect import GetSi, SmartConnect, Disconnect, GetStub, SoapStubAdapter
    from pyVmomi import vim, vmodl, VmomiSupport

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

log = logging.getLogger(__name__)


def get_config(config, profile=None, esxi_host=None):
    conf = (
        config.get("saltext.vmware")
        or config.get("grains", {}).get("saltext.vmware")
        or config.get("pillar", {}).get("saltext.vmware")
        or {}
    )
    if not conf:
        conf = (
            config.get("vmware_config")
            or config.get("grains", {}).get("vmware_config")
            or config.get("pillar", {}).get("vmware_config")
            or {}
        )
        if conf:
            log.warning(
                "Config found under vmware_config and not saltext.vmware. vmware_config has been deprecated and will be removed in 2023"
            )
    if profile:
        credentials = conf[profile]
    else:
        credentials = conf

    if esxi_host:
        host = esxi_host
        credentials = credentials.get("esxi_host", {}).get(esxi_host)
        password = credentials.get("password")
        user = credentials.get("user")
    else:
        host = os.environ.get("SALTEXT_VMWARE_HOST") or credentials.get("host")
        password = os.environ.get("SALTEXT_VMWARE_PASSWORD") or credentials.get("password")
        user = os.environ.get("SALTEXT_VMWARE_USER") or credentials.get("user")

    if host is None or password is None or user is None:
        raise ValueError("Cannot create service instance, VMware credentials incomplete.")
    return {"host": host, "user": user, "password": password}


def get_service_instance(*, config, esxi_host=None, profile=None):
    """
    Connect to VMware service instance.

    config
        The config to use to search for connection information. The search
        order matches that found in Salt's `config.get
        <https://docs.saltproject.io/en/latest/ref/modules/all/salt.modules.config.html#salt.modules.config.get>_`.
        Specifically the order is:

            1. Environment variables
            2. Minion config
            3. Minion grains
            4. Minion pillar data

        Environment variables are:
            SALTEXT_VMWARE_HOST
            SALTEXT_VMWARE_PASSWORD
            SALTEXT_VMWARE_USER
            SALTEXT_VMWARE_ESXI_USER
            SALTEXT_VMWARE_ESXI_PASSWORD

    esxi_host
        (optional) If specified, retrieves the configured username and password for the ESXi host.

    profile
        Profile to use (optional)

    Pillar Example:

    .. code-block:: yaml

        saltext.vmware:
            host: 198.51.100.100
            password: CorrectHorseBatteryStaple
            user: admin@example.com
            esxi_host:
                198.52.100.105:
                    user: admin
                    password: CorrectHorseBatteryStaple
                198.52.100.106:
                    user: admin
                    password: CorrectHorseBatteryStaple

    If configuration for multiple VMware services instances is required, they can be
    set up as different configuration profiles:

    .. code-block:: yaml

        saltext.vmware:
          profile1:
            host: 198.51.100.100
            password: CorrectHorseBatteryStaple
            user: admin@example.com
          profile2:
            host: 198.51.100.100
            password: CorrectHorseBatteryStaple
            user: admin@example.com
            esxi_host:
              198.52.100.105:
                user: admin
                password: CorrectHorseBatteryStaple
              198.52.100.106:
                user: admin
                password: CorrectHorseBatteryStaple

    """
    ctx = ssl._create_unverified_context()
    config = config or {}

    config = get_config(config=config, profile=profile, esxi_host=esxi_host)

    service_instance = connect.SmartConnect(
        host=config.get("host"),
        user=config.get("user"),
        pwd=config.get("password"),
        sslContext=ctx,
    )
    return service_instance


def request(url, method, body=None, token=None, opts=None, pillar=None):
    """
    Make a request to VMware rest api

    url
        url address for request.

    method
        Method for api request.

    body
        Body of the api request.

    token
        (optional) Api session token for api access, will create new token if not passed.

    opts
        (optional) Any additional options.

    pillar
        (optional) If specified, allows for a dictionary of pillar data to be made
        available to pillar and ext_pillar rendering. These pillar variables
        will also override any variables of the same name in pillar or
        ext_pillar.
    """
    # TODO: request needs test coverage -W. Werner, 2022-09-30
    host = (
        os.environ.get("SALTEXT_VMWARE_REST_API_HOST")
        or opts.get("saltext.vmware", {}).get("rest_api_host")
        or pillar.get("saltext.vmware", {}).get("rest_api_host")
        or os.environ.get("SALTEXT_VMWARE_HOST")
        or opts.get("saltext.vmware", {}).get("host")
        or pillar.get("saltext.vmware", {}).get("host")
    )
    cert = (
        os.environ.get("SALTEXT_VMWARE_REST_API_CERT")
        or opts.get("saltext.vmware", {}).get("rest_api_cert")
        or pillar.get("saltext.vmware", {}).get("rest_api_cert")
    )
    if not cert:
        cert = False
    if token is None:
        user = (
            os.environ.get("SALTEXT_VMWARE_REST_API_USER")
            or opts.get("saltext.vmware", {}).get("rest_api_user")
            or pillar.get("saltext.vmware", {}).get("rest_api_user")
            or os.environ.get("SALTEXT_VMWARE_USER")
            or opts.get("saltext.vmware", {}).get("user")
            or pillar.get("saltext.vmware", {}).get("user")
        )
        password = (
            os.environ.get("SALTEXT_VMWARE_REST_API_PASSWORD")
            or opts.get("saltext.vmware", {}).get("rest_api_password")
            or pillar.get("saltext.vmware", {}).get("rest_api_password")
            or os.environ.get("SALTEXT_VMWARE_PASSWORD")
            or opts.get("saltext.vmware", {}).get("password")
            or pillar.get("saltext.vmware", {}).get("password")
        )
        token = _get_session(host, user, password, cert)
    headers = {
        "Accept": "application/json",
        "content-Type": "application/json",
        "vmware-api-session-id": token,
    }
    session = requests.Session()
    response = session.request(
        method=method,
        url=f"https://{host}{url}",
        headers=headers,
        verify=cert,
        params=None,
        data=json.dumps(body),
    )
    return {"response": response, "token": token}


def _get_session(host, user, password, cert):
    """
    Create REST API session

    host
        Host for api request.

    user
        User to create session token for subsequent requests.

    password
        Password to create session token for subsequent requests.

    cert
        certificate for ssl verification.
    """
    headers = {"Accept": "application/json", "content-Type": "application/json"}
    session = requests.Session()
    if not cert:
        cert = False
    try:
        response = session.request(
            method="POST",
            url=f"https://{host}/rest/com/vmware/cis/session",
            headers=headers,
            auth=HTTPBasicAuth(user, password),
            verify=cert,
            params=None,
            data=json.dumps(None),
        )
        response.raise_for_status()
        json_response = response.json()
        return json_response["value"]

    except HTTPError as e:
        log.error(e)
        result = {"error": "Error occurred while calling vCenter API."}
        # if response contains json, extract error message from it
        if e.response.text:
            log.error(f"Response from vCenter {e.response.text}")
            try:
                error_json = e.response.json()
                if error_json["error_message"]:
                    result["error"] = e.response.json()["error_message"]
            except ValueError:
                log.error(
                    "Couldn't parse the response as json. Returning response text as error message"
                )
                result["error"] = e.response.text
        return result
    except SSLError as se:
        log.error(se)
        result = {"error": "SSL Error occurred while calling vCenter API."}
        return result
    except RequestException as re:
        log.error(re)
        result = {"error": "Error occurred while calling vCenter API."}
        return result



"""
Salt State file to create/update/delete transport nodes
Example:

.. code-block:: yaml

    create_transport_node:

      nsxt_transport_node.present:
        - name: Create Transport Node
             hostname: <nsxt-hostname>
             username: <nsxt-username>
             password: "<nsxt-password>
             validate_certs: False
             resource_type: "TransportNode"
             display_name: "NSX Configured TN"
             description: "NSX configured Test Transport Node"
             host_switch_spec:

                resource_type: "StandardHostSwitchSpec"
                required: False
                host_switches:
                - host_switch_profiles:

                  - name: "uplinkProfile1"
                    type: "UplinkHostSwitchProfile"

                host_switch_name: "hostswitch1"
                pnics:

                 - device_name: "vmnic1"
                   uplink_name: "uplink-1"

                ip_assignment_spec:
                  resource_type: "StaticIpPoolSpec"
                  ip_pool_name: "IPPool-IPV4-1"

                transport_zone_endpoints:
                - transport_zone_name: "TZ1"

             node_deployment_info:

               resource_type: "HostNode"
               display_name: "Host_1"
               ip_addresses: ["203.0.113.21"]
               os_type: "ESXI"
               os_version: "6.5.0"
               host_credential:

                 username: "root"
                 password: "ca$hc0w"
                 thumbprint: "e7fd7dd84267da10f991812ca62b2bedea3a4a62965396a04728da1e7f8e1cb9"
"""
import logging

import salt.utils.dictdiffer
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
from pyVmomi import vim

log = logging.getLogger(__name__)

try:
    from saltext.vmware.modules import nsxt_transport_node

    HAS_TRANSPORT_NODES = True
except ImportError:
    HAS_TRANSPORT_NODES = False


def __virtual__():
    if not HAS_TRANSPORT_NODES:
        return False, "'nsxt_transport_node' binary not found on system"
    return "nsxt_transport_node"


def _connect_with_vcentre(vcenter_host, username, password):
    vmware_config = {"host": vcenter_host, "user": username, "password": password}

    service_instance = connect.get_service_instance(config={"saltext.vmware": vmware_config})
    return service_instance


def inject_vcenter_info(hostname, username, password, ret, **kwargs):

    node_deployment_info = kwargs.get("node_deployment_info")
    if node_deployment_info:
        vm_deployment_config = node_deployment_info["deployment_config"]["vm_deployment_config"]

    if (
        vm_deployment_config
        and "vc_username" in vm_deployment_config
        and "vc_password" in vm_deployment_config
    ):
        vc_name = vm_deployment_config["vc_name"]

        vc_ip = _get_id_for_resource(
            hostname,
            username,
            password,
            "compute-manager",
            vc_name,
            ret,
            cert=kwargs.get("cert"),
            cert_common_name=kwargs.get("cert_common_name"),
            verify_ssl=kwargs.get("verify_ssl", True),
        )

        vc_username = vm_deployment_config.pop("vc_username", None)

        vc_password = vm_deployment_config.pop("vc_password", None)

        host = vm_deployment_config.pop("host", None)
        if host:
            vm_deployment_config["host_id"] = utils_common.get_mor_by_property(
                _connect_with_vcentre(vc_ip, vc_username, vc_password), vim.HostSystem, host
            )._moId

        storage = vm_deployment_config.pop("storage", None)
        if storage:
            vm_deployment_config["storage_id"] = utils_common.get_mor_by_property(
                _connect_with_vcentre(vc_ip, vc_username, vc_password), vim.Datastore, storage
            )._moId

        cluster = vm_deployment_config.pop("compute", None)
        if cluster:
            vm_deployment_config["compute_id"] = utils_common.get_mor_by_property(
                _connect_with_vcentre(vc_ip, vc_username, vc_password),
                vim.ClusterComputeResource,
                cluster,
            )._moId

        management_network = vm_deployment_config.pop("management_network", None)
        if management_network:
            vm_deployment_config["management_network_id"] = utils_common.get_mor_by_property(
                _connect_with_vcentre(vc_ip, vc_username, vc_password),
                vim.Network,
                management_network,
            )._moId

        data_networks = vm_deployment_config.pop("data_networks", None)
        if data_networks:
            vm_deployment_config["data_network_ids"] = utils_common.get_mor_by_property(
                _connect_with_vcentre(vc_ip, vc_username, vc_password), vim.Network, data_networks
            )._moId

        if "host" in vm_deployment_config:
            vm_deployment_config.pop("host", None)
        vm_deployment_config.pop("cluster", None)
        vm_deployment_config.pop("storage", None)
        vm_deployment_config.pop("management_network", None)
        vm_deployment_config.pop("data_networks", None)

    elif vm_deployment_config:
        if "host" in vm_deployment_config:
            host_id = vm_deployment_config.pop("host", None)
            vm_deployment_config["host_id"] = host_id

        cluster_id = vm_deployment_config.pop("compute", None)
        storage_id = vm_deployment_config.pop("storage", None)
        management_network_id = vm_deployment_config.pop("management_network", None)
        data_network_ids = vm_deployment_config.pop("data_networks", None)

        vm_deployment_config["compute_id"] = cluster_id
        vm_deployment_config["storage_id"] = storage_id
        vm_deployment_config["management_network_id"] = management_network_id
        vm_deployment_config["data_network_ids"] = data_network_ids
    else:
        return


# This method is called to form the payload of the create operation from the uer provided inputs
def _creation_of_request_body_for_tn(hostname, username, password, ret, **kwargs):
    node_deployment_info = kwargs.get("node_deployment_info")
    if node_deployment_info and node_deployment_info["resource_type"] == "EdgeNode":
        inject_vcenter_info(hostname, username, password, ret, **kwargs)

    _update_params_with_id(hostname, username, password, ret, **kwargs)
    return


def _get_id_for_resource(
    hostname,
    username,
    password,
    resource,
    resource_name,
    ret,
    cert=None,
    cert_common_name=None,
    verify_ssl=True,
):
    if resource == "ip_pool":
        get_ip_pool_response = __salt__["nsxt_ip_pools.get_by_display_name"](
            hostname,
            username,
            password,
            resource_name,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )
        if "error" in get_ip_pool_response:
            ret["result"] = False
            ret["comment"] = "Fail to get the IP Pool response : {}".format(
                get_ip_pool_response["error"]
            )
            return
        if len(get_ip_pool_response["results"]) > 1:
            ret["result"] = False
            ret["comment"] = "More than one IP Pool exist with same display name : {}".format(
                resource_name
            )
            return

        ip_pool_response_by_display_name = get_ip_pool_response["results"]
        ip_pool_dict = (
            ip_pool_response_by_display_name[0]
            if len(ip_pool_response_by_display_name) > 0
            else None
        )
        return ip_pool_dict["id"]

    elif resource == "transport_zone":
        get_transport_zone_response = __salt__["nsxt_transport_zone.get_by_display_name"](
            hostname,
            username,
            password,
            resource_name,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )
        if "error" in get_transport_zone_response:
            ret["result"] = False
            ret["comment"] = "Fail to get the transport zone response : {}".format(
                get_transport_zone_response["error"]
            )
            return
        if len(get_transport_zone_response["results"]) > 1:
            ret["result"] = False
            ret[
                "comment"
            ] = "More than one transport zone exist with same display name : {}".format(
                resource_name
            )
            return
        transport_zone_response_by_display_name = get_transport_zone_response["results"]
        transport_zone_dict = (
            transport_zone_response_by_display_name[0]
            if len(transport_zone_response_by_display_name) > 0
            else None
        )
        return transport_zone_dict["id"]

    elif resource == "host_switch_profile":

        get_host_switch_profile = __salt__["nsxt_uplink_profiles.get_by_display_name"](
            hostname,
            username,
            password,
            resource_name,
            include_system_owned=True,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )
        if "error" in get_host_switch_profile:
            ret["result"] = False
            ret["comment"] = "Fail to get the host switch profile response : {}".format(
                get_host_switch_profile["error"]
            )
            return
        if len(get_host_switch_profile["results"]) > 1:
            ret["result"] = False
            ret[
                "comment"
            ] = "More than one host switch profile exist with same display name : {}".format(
                resource_name
            )
            return
        host_profile_profile_by_display_name = get_host_switch_profile["results"]
        host_switch_profile_dict = (
            host_profile_profile_by_display_name[0]
            if len(host_profile_profile_by_display_name) > 0
            else None
        )
        return host_switch_profile_dict["id"]

    elif resource == "compute-manager":

        get_compute_manager_response = __salt__["nsxt_compute_manager.get_by_display_name"](
            hostname,
            username,
            password,
            resource_name,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )

        if "error" in get_compute_manager_response:
            ret["result"] = False
            ret["comment"] = "Fail to get the host compute manager resposne : {}".format(
                get_compute_manager_response["error"]
            )
            return
        if len(get_compute_manager_response["results"]) > 1:
            ret["result"] = False
            ret[
                "comment"
            ] = "More than one host compute managers exist with same display name : {}".format(
                resource_name
            )
            return
        compute_manager_by_display_name = get_compute_manager_response["results"]
        compute_manager_dict = (
            compute_manager_by_display_name[0] if len(compute_manager_by_display_name) > 0 else None
        )
        return compute_manager_dict["server"]

    elif resource == "compute-manager-id":

        get_compute_manager_response = __salt__["nsxt_compute_manager.get_by_display_name"](
            hostname,
            username,
            password,
            resource_name,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
        )

        if "error" in get_compute_manager_response:
            ret["result"] = False
            ret["comment"] = "Fail to get the host compute manager resposne : {}".format(
                get_compute_manager_response["error"]
            )
            return
        if len(get_compute_manager_response["results"]) > 1:
            ret["result"] = False
            ret[
                "comment"
            ] = "More than one host compute managers exist with same display name : {}".format(
                resource_name
            )
            return
        compute_manager_by_display_name = get_compute_manager_response["results"]
        compute_manager_dict = (
            compute_manager_by_display_name[0] if len(compute_manager_by_display_name) > 0 else None
        )
        return compute_manager_dict["id"]

    else:
        return None


def _update_params_with_id(hostname, username, password, ret, **kwargs):
    host_switch_spec = kwargs.get("host_switch_spec") or {}

    for host_switch in host_switch_spec.get("host_switches", {}):
        host_switch_profiles = host_switch.pop("host_switch_profiles", None)

        if host_switch_profiles:
            host_switch_profile_ids = []
            for host_switch_profile in host_switch_profiles:
                profile_obj = {}
                profile_obj["value"] = _get_id_for_resource(
                    hostname,
                    username,
                    password,
                    "host_switch_profile",
                    host_switch_profile["name"],
                    ret,
                    cert=kwargs.get("cert"),
                    cert_common_name=kwargs.get("cert_common_name"),
                    verify_ssl=kwargs.get("verify_ssl", True),
                )
                profile_obj["key"] = host_switch_profile["type"]
                host_switch_profile_ids.append(profile_obj)
                host_switch["host_switch_profile_ids"] = host_switch_profile_ids

        ip_assignment_spec = host_switch.get("ip_assignment_spec", {})
        if ip_assignment_spec.get("resource_type") == "StaticIpPoolSpec":
            ip_pool_name = ip_assignment_spec.pop("ip_pool_name", None)

            if ip_pool_name:
                ip_assignment_spec["ip_pool_id"] = _get_id_for_resource(
                    hostname,
                    username,
                    password,
                    "ip_pool",
                    ip_pool_name,
                    ret,
                    cert=kwargs.get("cert"),
                    cert_common_name=kwargs.get("cert_common_name"),
                    verify_ssl=kwargs.get("verify_ssl", True),
                )
        transport_zone_endpoints = host_switch.get("transport_zone_endpoints", {})

        for transport_zone_endpoint in transport_zone_endpoints:
            transport_zone_name = transport_zone_endpoint.pop("transport_zone_name", None)

            if transport_zone_name:
                transport_zone_endpoint["transport_zone_id"] = _get_id_for_resource(
                    hostname,
                    username,
                    password,
                    "transport_zone",
                    transport_zone_name,
                    ret,
                    cert=kwargs.get("cert"),
                    cert_common_name=kwargs.get("cert_common_name"),
                    verify_ssl=kwargs.get("verify_ssl", True),
                )

        # TODO support for logical-switches to be added once the end point is implemented

    return


def present(
    name,
    hostname,
    username,
    password,
    display_name,
    cert=None,
    cert_common_name=None,
    description=None,
    verify_ssl=None,
    host_switch_spec=None,
    node_deployment_info=None,
    transport_zone_endpoints=None,
    tags=None,
    maintenance_mode=None,
    is_overridden=None,
    resource_type=None,
):
    """
     Creates/Update(if present with the same name) of transport nodes
     .. code-block:: yaml

         create_transport_nodes:
           nsxt_transport_nodes.present:

             hostname: <nsxt-hostname>
             username: <nsxt-username>
             password: "<nsxt-password>
             validate_certs: False
             resource_type: "TransportNode"
             display_name: "NSX Configured TN"
             description: "NSX configured Test Transport Node"
             host_switch_spec:

                resource_type: "StandardHostSwitchSpec"
                required: False
                host_switches:
                - host_switch_profiles:

                  - name: "uplinkProfile1"
                    type: "UplinkHostSwitchProfile"

                host_switch_name: "hostswitch1"
                pnics:

                 - device_name: "vmnic1"
                   uplink_name: "uplink-1"

                ip_assignment_spec:
                  resource_type: "StaticIpPoolSpec"
                  ip_pool_name: "IPPool-IPV4-1"

                transport_zone_endpoints:
                - transport_zone_name: "TZ1"

             node_deployment_info:

               resource_type: "HostNode"
               display_name: "Host_1"
               ip_addresses: ["203.0.113.21"]
               os_type: "ESXI"
               os_version: "6.5.0"
               host_credential:

                 username: "root"
                 password: "ca$hc0w"
                 thumbprint: "e7fd7dd84267da10f991812ca62b2bedea3a4a62965396a04728da1e7f8e1cb9"

    hostname
        The host name of NSX-T manager
    username
        Username to connect to NSX-T manager
    password
        Password to connect to NSX-T manager
    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.
    cert
        Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.
    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against
    transport-node-id
        Id of the transport node which needs to be updated
    revision
        revision number of the transport node
    node_deployment_revision
        revision number of the node deployment number which was part of the body
    description
        Description of transport nodes
    host_switch_spec:
        description: This property is used to either create standard host switches
              or to inform NSX about preconfigured host switches that already
              exist on the transport node.
              Pass an array of either StandardHostSwitchSpec objects or
              PreconfiguredHostSwitchSpec objects. It is an error to pass
              an array containing different types of HostSwitchSpec objects.'
        host_switches: This property is deprecated in favor of 'host_switch_spec'. Property
                  'host_switches' can only be used for NSX managed transport nodes.
                  'host_switch_spec' can be used for both NSX managed or manually
                  preconfigured host switches.

        resource_type: Selects the type of the transport zone profile
    maintenance_mode:
        description: The property is read-only, used for querying result. User could update
                  transport node maintenance mode by UpdateTransportNodeMaintenanceMode call.
    node_deployment_info:
        Allocation_list: Array of logical router ids to which this edge node is allocated.
    deployment_config:
        description: 'When this configuration is specified, edge fabric node of deployment_type
                      VIRTUAL_MACHINE will be deployed and registered with MP.'

        form_factor:
        node_user_settings:

            audit_password:
                description: "Password for the node audit user. For deployment, this property
                              is required. After deployment, this property is ignored, and
                              the node cli must be used to change the password. The password
                              specified must be at least 12 characters in length and must
                              contain at least one lowercase, one uppercase, one numeric
                              character and one special character (except quotes)."
            audit_username:
               description: "The default username is \"audit\". To configure username, you
                             must provide this property together with <b>audit_password</b>."
            cli_password:
               description: "Password for the node cli user. For deployment, this property
                          is required. After deployment, this property is ignored, and
                          the node cli must be used to change the password. The password
                          specified must be at least 12 characters in length and must
                          contain at least one lowercase, one uppercase, one numeric
                          character and one special character (except quotes)."
            cli_username:
               description: "To configure username, you must provide this property together
                             with <b>cli_password</b>."
            description: "Username and password settings for the node. Note - these settings
                         will be honored only during node deployment. Post deployment, CLI
                         must be used for changing the user settings, changes to these
                         parameters will not have any effect."
            root_password:
               description: "Password for the node root user. For deployment, this property
                            is required. After deployment, this property is ignored, and the
                            node cli must be used to change the password. The password
                            specified must be at least 12 characters in length and must
                            contain at least one lowercase, one uppercase, one numeric
                            character and one special character (except quotes)."

        vm_deployment_config:
            allow_ssh_root_login:
               description: 'If true, the root user will be allowed to log into the VM.
                             Allowing root SSH logins is not recommended for security
                             reasons.'
            compute:
                description: 'The cluster node VM will be deployed on the specified cluster
                              or resourcepool for specified VC server. If vc_username and
                              vc_password are present then this field takes name else id.'
            data_networks:
                description: "List of distributed portgroup or VLAN logical identifiers or names to
                   which the datapath serving vnics of edge node vm will be connected. If vc_username
                   and vc_password are present then this field takes names else id."
            default_gateway_addresses:
                description: 'The default gateway for the VM to be deployed must be specified
                              if all the other VMs it communicates with are not in the same subnet.
                              Do not specify this field and management_port_subnets to use DHCP.
                              Note: only single IPv4 default gateway address is supported and it
                              must belong to management network.
                              IMPORTANT: VMs deployed using DHCP are currently not supported,
                              so this parameter should be specified.'

            description: VM Deployment Configuration
            dns_servers:

                description: 'List of DNS servers
                              If DHCP is used, the default DNS servers associated with
                              the DHCP server will be used instead.
                              Required if using static IP.'

            enable_ssh:
                description: 'If true, the SSH service will automatically be started on the
                              VM. Enabling SSH service is not recommended for security
                              reasons.'

            host:
                description: "Name of the host where edge VM is to be deployed
                              if vc_username and vc_password are present then
                              this field takes host name else host id."

            hostname:
                description: Desired host name/FQDN for the VM to be deployed

            management_network:
                description: 'Distributed portgroup identifier to which the management vnic
                              of cluster node VM will be connected. If vc_username and vc_password
                              are present then this field takes name else id.'

            management_port_subnets:
                description: 'IP Address and subnet configuration for the management port.
                              Do not specify this field and default_gateway_addresses to
                              use DHCP.
                              Note: only one IPv4 address is supported for the management
                              port.
                              IMPORTANT: VMs deployed using DHCP are currently not supported,
                              so this parameter should be specified.'

            ntp_servers:
                description: 'List of NTP servers.
                  To use hostnames, a DNS server must be defined. If not using DHCP,
                  a DNS server should be specified under dns_servers.'
            placement_type:
                description: "Specifies the config for the platform through which to deploy
                   the VM"
            search_domains:
                description: 'List of domain names that are used to complete unqualified host
                              names.'
            storage:
                description: Moref or name of the datastore in VC. If it is to be taken from 'Agent
                             VM Settings', then it should be empty If vc_username and vc_password are present then
                             this field takes name else id.
            vc_name:
                description: 'The VC-specific names will be resolved on this VC, so all
                              other identifiers specified in the config must belong to this vCenter
                              server.'
            vc_username:
                description: 'Username of VC'

            vc_password:
                description: 'VC Password'

            reservation_info:
                description: 'Resource reservation for memory and CPU resources'
                cpu_reservation:

                    description: 'Guaranteed minimum allocation of CPU resources'
                    reservation_in_mhz:

                        description: 'GCPU resevation in mhz'

                    reservation_in_shares:

                        description: 'CPU reservation in shares'

                memory_reservation:
                    description: 'Guaranteed minimum allocation of memory resources'
                    reservation_percentage:

                        description: 'Memory reservation percentage'

            resource_allocation:
                description: 'Resource reservation settings'
                cpu_count:

                    description: 'CPU count'

                memory_allocation_in_mb:

                    description: 'Memory allocation in MB'

    deployment_type:
        description: Specifies whether the service VM should be deployed on each host such
                     that it provides partner service locally on the host, or whether the
                     service VMs can be deployed as a cluster. If deployment_type is
                     CLUSTERED, then the clustered_deployment_count should be provided.
    discovered_ip_addresses:
        description: Discovered IP Addresses of the fabric node, version 4 or 6
    discovered_node_id:
        description: Id of discovered node which was converted to create this node
    external_id:
         description: Current external id of this virtual machine in the system.
    fqdn:
        description: Domain name the entity binds to
    host_credential:
        description: Login credentials for the host
        password:

            description: Password for the user (optionally specified on PUT, unspecified
             on GET)

        thumbprint:
            description: Hexadecimal SHA256 hash of the vIDM server's X.509 certificate
        username:
            description: Username value of the log
    ip_addresses:
        description: Interface IP addresses
    managed_by_server:
        description: The id of the vCenter server managing the ESXi type HostNode
    os_type:
        description: OS type of the discovered node
    os_version:
        description: OS version of the discovered node
    resource_type:
        description: Selects the type of the transport zone profile
    remote_tunnel_endpoint:
       description: Configuration for a remote tunnel endpoin
       host_switch_name:

           description: The host switch name to be used for the remote tunnel endpoint

       named_teaming_policy:

           description: The named teaming policy to be used by the remote tunnel endpoint

       rtep_vlan:

           description: VLAN id for remote tunnel endpoint

       VlanID:

           description: Virtual Local Area Network Identifier

       ip_assignment_spec:

           description: Specification for IPs to be used with host switch remote tunnel endpoints
           resource_type:

               description: Resource type

           ip_pool_id:

               description: IP pool id

           ip_list:

               description: List of IPs for transport node host switch virtual tunnel endpoints

           ip_mac_list:

               description: List of IPs and MACs for transport node host switch virtual tunnel endpoints

           default_gateway:
               description: Default gateway
               IPAddress:

                   description: IPv4 or IPv6 address

           subnet_mask:
               description: Subnet mask
               IPAddress:

                   description: IPv4 IPv6 address
    tags:
        description: Opaque identifiers meaningful to the API user

    transport_zone_endpoints:
       description: Transport zone endpoints.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    get_transport_node = __salt__["nsxt_transport_node.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        display_name=display_name,
        cert=cert,
        verify_ssl=verify_ssl,
        cert_common_name=cert_common_name,
    )

    transport_node_id = revision = node_deployment_revision = None
    is_update = False

    if "error" in get_transport_node:
        ret["result"] = False
        ret["comment"] = "Failed to get the transport nodes : {}".format(
            get_transport_node["error"]
        )
        return ret

    transport_node_response_by_display_name = get_transport_node["results"]

    if len(transport_node_response_by_display_name) > 1:
        ret["result"] = True
        ret["comment"] = "More than one transport node exist with same display name : {}".format(
            display_name
        )
        return ret

    transport_node_dict = (
        transport_node_response_by_display_name[0]
        if transport_node_response_by_display_name
        else None
    )

    if transport_node_dict:
        transport_node_id = transport_node_dict["id"]
        revision = transport_node_dict["_revision"]
        is_update = salt.utils.dictdiffer.deep_diff(
            transport_node_dict, {"description": description, "host_switch_spec": host_switch_spec}
        )
        if "node_deployment_info" in transport_node_dict:
            node_deployment_revision = transport_node_dict["node_deployment_info"]["_revision"]

    if __opts__.get("test"):
        log.info("present is called with test option")
        if transport_node_dict is None:
            ret["result"] = None
            ret["comment"] = "Transport Node will be created in NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "Transport Node will be updated"
        return ret

    _creation_of_request_body_for_tn(
        hostname,
        username,
        password,
        ret,
        cert=cert,
        verify_ssl=verify_ssl,
        cert_common_name=cert_common_name,
        node_deployment_info=node_deployment_info,
        host_switch_spec=host_switch_spec,
    )

    if ret["result"] == False:
        return ret

    if is_update:
        log.info("Start of the update of the transport nodes")
        update_transport_node = __salt__["nsxt_transport_node.update"](
            hostname=hostname,
            username=username,
            password=password,
            transport_node_id=transport_node_id,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
            node_deployment_revision=node_deployment_revision,
            revision=revision,
            description=description,
            display_name=display_name,
            node_deployment_info=node_deployment_info,
            host_switch_spec=host_switch_spec,
            tags=tags,
            maintenance_mode=maintenance_mode,
            is_overridden=is_overridden,
            transport_zone_endpoints=transport_zone_endpoints,
            resource_type=resource_type,
        )

        if "error" in update_transport_node:
            ret["result"] = False
            ret["comment"] = "Fail to update transport_node : {}".format(
                update_transport_node["error"]
            )
            return ret

        ret["comment"] = "Transport node updated successfully"
        ret["result"] = True
        ret["changes"]["old"] = transport_node_dict
        ret["changes"]["new"] = update_transport_node
        return ret
    else:
        if transport_node_id is not None:
            ret["result"] = True
            ret["comment"] = "Transport node with display_name %s already exists", display_name
            return ret

        else:
            log.info("Start of the create of the transport node")
            create_transport_node = __salt__["nsxt_transport_node.create"](
                hostname=hostname,
                username=username,
                password=password,
                cert=cert,
                cert_common_name=cert_common_name,
                display_name=display_name,
                description=description,
                verify_ssl=verify_ssl,
                transport_zone_endpoints=transport_zone_endpoints,
                host_switch_spec=host_switch_spec,
                node_deployment_info=node_deployment_info,
                tags=tags,
                maintenance_mode=maintenance_mode,
                is_overridden=is_overridden,
                resource_type=resource_type,
            )

        if "error" in create_transport_node:
            ret["result"] = False
            ret["comment"] = "Fail to create transport_node : {}".format(
                create_transport_node["error"]
            )
            return ret

        ret["result"] = True
        ret["comment"] = "Transport node created successfully"
        ret["changes"]["new"] = create_transport_node
        return ret


def absent(
    name,
    hostname,
    username,
    password,
    display_name,
    verify_ssl=None,
    cert=None,
    cert_common_name=None,
):
    """
    Deletes an Transport Nodes of the of provided name (if present)
    .. code-block:: yaml

        delete_ip_pool:
          nsxt_transport_node.absent:

            - name: transport node delete
              hostname: <hostname>
              username: <username>
              password: <password>
              certificate: <certificate>
              verify_ssl: <False/True>
              display_name: <ip pool name>

    hostname
        The host name of NSX-T manager
    username
        Username to connect to NSX-T manager
    password
        Password to connect to NSX-T manager
    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.
    cert
        Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.
    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against
    display_name
        Display name of transport node to delete
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    get_transport_node = __salt__["nsxt_transport_node.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        display_name=display_name,
        cert=cert,
        verify_ssl=verify_ssl,
        cert_common_name=cert_common_name,
    )

    transport_node_id = None

    if "error" in get_transport_node:
        ret["result"] = False
        ret["comment"] = "Failed to get the transport nodes : {}".format(
            get_transport_node["error"]
        )
        return ret

    transport_node_response_by_display_name = get_transport_node["results"]

    if len(transport_node_response_by_display_name) > 1:
        ret["result"] = True
        ret["comment"] = "More than one transport node exist with same display name : {}".format(
            display_name
        )
        return ret

    transport_node_dict = (
        transport_node_response_by_display_name[0]
        if len(transport_node_response_by_display_name) > 0
        else None
    )

    if transport_node_dict is not None:
        transport_node_id = transport_node_dict["id"]
    else:
        ret["result"] = True
        ret["comment"] = "No transport node exist with display name %s" % display_name
        return ret

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if transport_node_dict is not None:
            ret["result"] = None
            ret["comment"] = "Transport Node will be deleted in NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "State absent will do nothing , since transport node is not existing"
        return ret

    delete_response = __salt__["nsxt_transport_node.delete"](
        hostname=hostname,
        username=username,
        password=password,
        transport_node_id=transport_node_id,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in delete_response:
        ret["result"] = False
        ret["comment"] = "Failed to delete the transport-node : {}".format(delete_response["error"])
        return ret

    ret["changes"]["old"] = transport_node_dict
    ret["changes"]["new"] = {}
    ret["comment"] = "Transport node deleted successfully"
    ret["result"] = True
    return ret



"""
NSX-T Datacenter Configuration Management state module
======================================================

:maintainer: <VMware>
:maturity: new

Enable/Disable the NSX-T manager's publish_fqdns status.

Example usage:

.. code-block:: yaml

    configure_nsx_manager:
      nsx_manager.publish_fqdns_enabled:
        - name: Enable publish fqdns
          hostname: <hostname>
          username: <username>
          password: <password>
          cert: <certificate pem file path>
          verify_ssl: <False/True>

.. warning::

    It is recommended to pass the NSX authentication details using Pillars rather than specifying as plain text in SLS
    files.

"""
import logging

log = logging.getLogger(__name__)

__virtual_name__ = "nsxt_manager"


def __virtual__():
    """
    Only load if the nsxt_manager module is available in __salt__
    """
    return (
        __virtual_name__ if "nsxt_manager.get_manager_config" in __salt__ else False,
        "'nsxt_manager' binary not found on system",
    )


def _set_publish_fqdns_in_nsxt(
    publish_fqdns, revision, hostname, username, password, verify_ssl, cert, cert_common_name
):
    out = __salt__["nsxt_manager.set_manager_config"](
        hostname=hostname,
        username=username,
        password=password,
        revision=revision,
        publish_fqdns=publish_fqdns,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )
    return out


def _get_publish_fqdns_revision_from_nsxt(
    hostname, username, password, verify_ssl, cert, cert_common_name
):
    out = __salt__["nsxt_manager.get_manager_config"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )
    return out


def _get_publish_fqdns_revision_from_response(current_config_response):
    current_publish_fqdns = current_config_response.get("publish_fqdns")
    current_revision = current_config_response.get("_revision")

    return current_publish_fqdns, current_revision


def _create_state_response(name, old_state, new_state, result, comment):
    state_response = dict()
    state_response["name"] = name
    state_response["result"] = result
    state_response["comment"] = comment
    state_response["changes"] = dict()
    if new_state:
        state_response["changes"]["old"] = old_state
        state_response["changes"]["new"] = new_state

    return state_response


def publish_fqdns_enabled(
    name, hostname, username, password, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Check the value for publish_fqdns, if it is true then do nothing else set it to true

    .. code-block:: yaml

        configure_nsxt_manager:
          nsx_manager.publish_fqdns_enabled:
            - name: Enable publish fqdns
            - hostname: <hostname>
            - username: <username>
            - password: <password>
            - cert: <certificate path>
            - verify_ssl: <False/True>

    name
        The Operation to perform

    hostname
        NSX-T manager's hostname

    username
        NSX-T manager's username

    password
        NSX-T manager's password

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """

    log.info("Getting the manager's config")

    cert_common_name = cert_common_name
    cert = cert

    get_current_config = _get_publish_fqdns_revision_from_nsxt(
        hostname,
        username,
        password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )
    if "error" in get_current_config:
        return _create_state_response(name, None, None, False, get_current_config["error"])

    current_publish_fqdns, current_revision = _get_publish_fqdns_revision_from_response(
        get_current_config
    )

    if __opts__.get("test"):
        log.info("publish_fqdns_enabled is called with test option")
        return _create_state_response(
            name,
            None,
            None,
            None,
            "State publish_fqdns_enabled will execute with params {}, {}, {}, {}, {}".format(
                name, hostname, username, password, verify_ssl
            ),
        )

    if current_publish_fqdns:
        return _create_state_response(
            name, None, None, True, "publish_fqdns is already set to True"
        )

    publish_fqdns = True

    log.info("Updating the NSX-T manager's config")
    updated_config_response = _set_publish_fqdns_in_nsxt(
        publish_fqdns,
        current_revision,
        hostname,
        username,
        password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in updated_config_response:
        return _create_state_response(name, None, None, False, updated_config_response["error"])

    return _create_state_response(
        name,
        get_current_config,
        updated_config_response,
        True,
        "publish_fqdns has been set to True",
    )


def publish_fqdns_disabled(
    name, hostname, username, password, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Check the value for publish_fqdns, if it is false then do nothing else set it to false

    .. code-block:: yaml

        configure_nsxt_manager:
          nsx_manager.publish_fqdns_disabled:
            - name: Disable publish fqdns
            - hostname: <hostname>
            - username: <username>
            - password: <password>
            - cert: <certificate>
            - verify_ssl: <False/True>

    name
        The Operation to perform

    hostname
        NSX-T manager's hostname

    username
        NSX-T manager's username

    password
        NSX-T manager's password

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """

    cert_common_name = cert_common_name
    cert = cert

    log.info("Getting the manager's config")
    get_current_config = _get_publish_fqdns_revision_from_nsxt(
        hostname,
        username,
        password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )
    if "error" in get_current_config:
        return _create_state_response(name, None, None, False, get_current_config["error"])

    current_publish_fqdns, current_revision = _get_publish_fqdns_revision_from_response(
        get_current_config
    )

    if __opts__.get("test"):
        log.info("publish_fqdns_disabled is called with test option")
        return _create_state_response(
            name,
            None,
            None,
            None,
            "State publish_fqdns_disabled will execute with params {}, {}, {}, {}, {}".format(
                name, hostname, username, password, verify_ssl
            ),
        )

    if not current_publish_fqdns:
        return _create_state_response(
            name, None, None, True, "publish_fqdns is already set to False"
        )

    publish_fqdns = False

    log.info("Updating the manager's config")
    updated_config_response = _set_publish_fqdns_in_nsxt(
        publish_fqdns,
        current_revision,
        hostname,
        username,
        password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in updated_config_response:
        return _create_state_response(name, None, None, False, updated_config_response["error"])

    return _create_state_response(
        name,
        get_current_config,
        updated_config_response,
        True,
        "publish_fqdns has been set to False",
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
# pylint: disable=no-name-in-module
import logging

# noreorder
import salt.exceptions
import salt.modules.cmdmod
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import saltext.vmware.utils.cluster as utils_cluster
import saltext.vmware.utils.datacenter as utils_datacenter
import saltext.vmware.utils.esxi as utils_esxi
from saltext.vmware.utils.common import get_service_content as get_service_content

try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

log = logging.getLogger(__name__)


def _list_lic_keys(licenses):
    """
    Return a list of license keys

    licenses:
       vCenter licenses as made available from licenseManager

    Note:
        skips evaluation licenses
    """
    lic_keys = []
    for lic in licenses:
        if lic.used is None:
            continue
        lic_keys.append(lic.licenseKey)
    return lic_keys


def _find_lic_for_key(licenses, license_key):
    """
    Return license which matches specified license_key

    licenses:
       vCenter licenses as made available from licenseManager

    license_key
        license key to search for

    Note:
        skips evaluation licenses
    """
    lic_keys = []
    for lic in licenses:
        if lic.used is None:
            continue
        if license_key == lic.licenseKey:
            return lic
    return None


def _get_entity_id(service_instance, datacenter_name, cluster_name, esxi_hostname):
    """
    Get entity_id for datacenter, cluster, ESXi Server or vCenter

    If no datacenter, cluster or ESXi Server is specified, it is assumed the operation is to be applied to a vCenter

    service_instance:
        Service Instance

    datacenter_name
        Datacenter name to use for the operation

    cluster_name
        Name of the cluster to use for the operation

    esxi_hostname
        Hostname of the ESXi Server use for the operation
    """

    # need to extract entity this license is intended for
    # from choices:
    #   1 vCenter - _entity_id = service_instance.content.about.instanceUuid
    #   2 ESXi - _entity_id = esxi host _moId
    #   3 cluster - _entity_id = cluster _moId

    _entity_id = None  # TBD miracle happens here and have value
    datacenter_ref = None
    if not any((datacenter_name, cluster_name, esxi_hostname)):
        # default to applying to vCenter
        srv_content = get_service_content(service_instance)
        _entity_id = srv_content.about.instanceUuid
    else:
        if datacenter_name:
            # need to get named datacenter's reference
            datacenter_ref = utils_datacenter.get_datacenter(service_instance, datacenter_name)
            log.debug(
                f"retrieved datacenter ref '{datacenter_ref }' for datacenter '{datacenter_name}'"
            )

        if cluster_name:
            # need to get named cluster's reference
            cluster_ref = utils_cluster.get_cluster(dc_ref=datacenter_ref, cluster=cluster_name)
            _entity_id = cluster_ref._moId
            log.debug(
                f"retrieved entityId '{_entity_id}' from cluster ref '{cluster_ref }' for cluster '{cluster_name}' and datacenter '{datacenter_name}'"
            )

        if esxi_hostname:
            # need to get named esxi server
            esxi_hosts = utils_esxi.get_hosts(service_instance, datacenter_name, esxi_hostname)

            # returns hosts list of dicts
            if not esxi_hosts:
                log.debug(f"Failed to find esxi hostname '{esxi_hostname}'")
                return False

            if len(esxi_hosts) > 1:
                log.error(
                    f"Failed, found multiple instances of esxi hostname '{esxi_hostname}', hosts returned '{esxi_hosts}'"
                )
                return False

            _entity_id = esxi_hosts[0]._moId

    return _entity_id


def is_vcenter(service_instance):
    """
    Test if service_instance represents vCenter,

    service_instance
        The Service Instance to check if it represents a vCenter

    Return:
        True  - vCenter
        False - not a vCenter
    """
    try:
        srv_content = get_service_content(service_instance)
        apitype = srv_content.about.apiType
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    except AttributeError as exc:
        log.exception(exc)
        raise salt.exceptions.CommandExecutionError(exc)
    except TypeError as exc:
        log.exception(exc)
        raise salt.exceptions.CommandExecutionError(exc)

    if apitype == "VirtualCenter":
        return True

    return False


def get_license_mgr(service_instance):
    """
    Return a license manager from specified Service Instance
    if the Service Instance is connected to a vCenter,
    otherwise return None.

    service_instance
        vCenter service connection instance

    Note: returns LicenseManager
    """
    if not is_vcenter(service_instance):
        return None

    srv_content = get_service_content(service_instance)
    return srv_content.licenseManager


def get_license_assignment_mgr(service_instance):
    """
    Return a license assignment manager from specified Service Instance
    if the Service Instance is connected to a vCenter,
    otherwise return None

    service_instance
        vCenter service connection instance

    Note: returns LicenseAssignmentManager
    """
    if not is_vcenter(service_instance):
        return None

    srv_content = get_service_content(service_instance)
    return srv_content.licenseManager.licenseAssignmentManager


def list_licenses(service_instance):
    """
    Returns a list of licenses associated with specified Service Instance.

    service_instance
        The Service Instance Object from which to obtain licenses.
    """
    ret = {}
    lic_mgr = get_license_mgr(service_instance)
    log.debug(f"License Manager listing of licenses '{lic_mgr.licenses}'")

    if not lic_mgr:
        ret["message"] = "Failed, not connected to a vCenter"
        ret["result"] = False
        return ret

    ret["licenses"] = _list_lic_keys(lic_mgr.licenses)
    return ret


def add_license(
    service_instance, license_key, datacenter_name=None, cluster_name=None, esxi_hostname=None
):
    """
    Add license to the pool of available licenses associated with the License Manager
    and assign it to the specified Cluster, ESXi Server or vCenter

    If no datacenter, cluster or ESXi Server is specified, it is assumed the operation is to be applied to a vCenter

    service_instance:
        Service Instance containing a License Manager

    license_key
        License key which specifies license to add to license manager

    datacenter_name
        Datacenter name to use for the operation [default None]

    cluster_name
        Name of the cluster to add license [default None]

    esxi_hostname
        Hostname of the ESXi Server to add license [default None]

    Returns:
        True - if successful or license already present
    """
    lic_mgr = get_license_mgr(service_instance)
    if not lic_mgr:
        return False

    lic_keys = _list_lic_keys(lic_mgr.licenses)
    log.debug(f"attempting to add license key to list of existing license keys '{lic_keys}'")

    try:
        if not license_key in lic_keys:
            licMgrLicInfo = lic_mgr.AddLicense(licenseKey=license_key)
            log.debug(f"attempted AddLicense, returned license into '{licMgrLicInfo}'")
            props_list = []
            for prop in licMgrLicInfo.properties:
                props_list.append({prop.key: prop.value})

            # check for failure to add license
            license_add_failure = False
            license_add_failure_msg = ""
            for prop_dict in props_list:
                if "lc_error" in prop_dict:
                    log.error(f"Failed AddLicense for license key")
                    license_add_failure = True
                if "diagnostic" in prop_dict:
                    license_add_failure_msg = prop_dict["diagnostic"]
            if license_add_failure:
                log.error(f"Failed to add license key, reason '{license_add_failure_msg}'")
                raise salt.exceptions.CommandExecutionError(license_add_failure_msg)

        # get license just added for specified license key
        addedLic = _find_lic_for_key(lic_mgr.licenses, license_key)
        if not addedLic:
            log.error(f"Unable to find license for recently added license_key")
            return False

        entity_id = _get_entity_id(service_instance, datacenter_name, cluster_name, esxi_hostname)
        lic_assign_mgr = get_license_assignment_mgr(service_instance)
        if entity_id:
            assigned_lic = lic_assign_mgr.QueryAssignedLicenses(entityId=entity_id)

            log.debug(
                f"assigning license, entity identifier '{entity_id}' has assigned license '{assigned_lic}'"
            )
            # pyVmomi seen doing strange things, hence checking length returned
            if not assigned_lic or (
                len(assigned_lic) != 0 and assigned_lic[0].assignedLicense.licenseKey != license_key
            ):
                lic_assign_mgr.UpdateAssignedLicense(entity=entity_id, licenseKey=license_key)

    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    except AttributeError as exc:
        log.exception(exc)
        raise salt.exceptions.CommandExecutionError(exc)
    except TypeError as exc:
        log.exception(exc)
        raise salt.exceptions.CommandExecutionError(exc)

    return True


def remove_license(
    service_instance, license_key, datacenter_name=None, cluster_name=None, esxi_hostname=None
):
    """
    Remove license from the pool of available licenses associated with the License Manager
    and unassign it from the specified Cluster, ESXi Server or vCenter

    If no datacenter, cluster or ESXi Server is specified, it is assumed the operation is to be applied to a vCenter

    service_instance:
        Service Instance containing a License Manager

    license_key
        License key which specifies license to remove from the license manager

    datacenter_name
        Datacenter name to use for the operation [default None]

    cluster_name
        Name of the cluster to add license [default None]

    esxi_hostname
        Hostname of the ESXi Server to add license [default None]

    Returns:
        True - if successful, license removed
    """
    lic_mgr = get_license_mgr(service_instance)
    if not lic_mgr:
        log.debug(f"attempting to remove license but unable to find license manager")
        return False

    lic_keys = _list_lic_keys(lic_mgr.licenses)
    log.debug(f"attempting to remove license from list of existing license keys '{lic_keys}'")
    if license_key not in lic_keys:
        log.debug(
            f"cannot remove license key since not found in list of existing license keys '{lic_keys}'"
        )
        return False

    # Need to fine where license key is assigned and then un-assign it,
    # and only then can the license actually be removed, that is, RemoveLicense is a noop
    # if the license is assigned
    #
    # RemoveAssignedLicense takes entityId, reason to find were license key is assigned
    # Note: an entity can only have one license assigned

    try:
        entity_id = _get_entity_id(service_instance, datacenter_name, cluster_name, esxi_hostname)
        lic_assign_mgr = get_license_assignment_mgr(service_instance)
        if entity_id:
            assigned_lic = lic_assign_mgr.QueryAssignedLicenses(entityId=entity_id)

            log.debug(
                f"assigning license, entity identifier '{entity_id}' has assigned license '{assigned_lic}'"
            )
            if assigned_lic and assigned_lic[0].assignedLicense.licenseKey == license_key:
                log.debug(f"Unassigning license key from entity identifer '{entity_id}'")
                lic_assign_mgr.RemoveAssignedLicense(entityId=entity_id)

            else:
                log.debug(
                    f"no assigned license found, or the assigned license key for entity identifier '{entity_id}' did not match specified license key"
                )

            log.debug(f"Removing license key from License Managers pool of licenses")
            lic_mgr.RemoveLicense(licenseKey=license_key)
        else:
            log.debug(
                f"Unable to find entity for specified inputs when attempting to remove license key"
            )
            return False

    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    except AttributeError as exc:
        log.exception(exc)
        raise salt.exceptions.CommandExecutionError(exc)
    except TypeError as exc:
        log.exception(exc)
        raise salt.exceptions.CommandExecutionError(exc)

    return True



"""
State module for NSX-T compute manager registration and de-registration
"""
import logging

log = logging.getLogger(__name__)

try:
    from saltext.vmware.modules import nsxt_compute_manager

    HAS_NSXT_COMPUTE_MANAGER = True
except ImportError:
    HAS_NSXT_COMPUTE_MANAGER = False


def __virtual__():
    if not HAS_NSXT_COMPUTE_MANAGER:
        return False, "'nsxt_compute_manager' binary not found on system"
    return "nsxt_compute_manager"


def _needs_update(
    existing_compute_manager, thumbprint, server_origin_type, **compute_manager_params
):
    if thumbprint and thumbprint != existing_compute_manager["credential"]["thumbprint"]:
        return True
    if server_origin_type and server_origin_type != existing_compute_manager["origin_type"]:
        return True

    optional_params = ("display_name", "description", "set_as_oidc_provider")
    for param in optional_params:
        param_val = compute_manager_params.get(param)
        if param_val and existing_compute_manager[param] != param_val:
            return True
    return False


def present(
    name,
    hostname,
    username,
    password,
    compute_manager_server,
    credential,
    server_origin_type="vCenter",
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    display_name=None,
    description=None,
    set_as_oidc_provider=None,
):
    """

    Registers compute manager in NSX-T Manager or updates the registration

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.present hostname=nsxt-manager.local username=admin ...

    .. code-block:: yaml

        register_compute_manager:
          nsxt_compute_manager.present:
            - name: Registration
              hostname: {{ pillar['nsxt_manager_hostname'] }}
              username: {{ pillar['nsxt_manager_username'] }}
              password: {{ pillar['nsxt_manager_password'] }}
              cert: {{ pillar['nsxt_manager_certificate'] }}
              verify_ssl: <False/True>
              compute_manager_server: <compute manager IP address or FQDN>
              server_origin_type: <compute manager origin type>
              credential:
                credential_type:  UsernamePasswordLoginCredential
                username: {{ pillar['compute manager username'] }}
                password: {{ pillar['compute manager password'] }}
                thumbprint: {{ pillar['compute manager thumbprint'] }}
              display_name: <compute manager name>
              description: <compute manager description>
              set_as_oidc_provider: <False/True>

    name
        name of the operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    compute_manager_server
        Compute manager server IP address or FQDN

    credential
        An object which contains credential details to validate compute manager
        Sample usage in sls file:

        .. code::

            credential:
               credential_type: "UsernamePasswordLoginCredential"
               username: "user"
               password: "pass"
               thumbprint: "36:XX:XX:XX:XX:XX:XX66"

        credential_type
            Type of credential provided. For now only UsernamePasswordLoginCredential is supported.

        username
            Username of the compute manager

        password
            Password of the compute manager

        thumbprint
            Thumbprint of the compute manager

    server_origin_type
        (Optional) Server origin type of the compute manager. Default is vCenter

    display_name
        (Optional) Display name of the compute manager

    description
        (Optional) description for the compute manager

    set_as_oidc_provider
        (Optional) Specifies whether compute manager has been set as OIDC provider. Default is false. If the
         compute manager is VC and need to set set as OIDC provider for NSX then this flag should be set as true.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    For more information, see `VMware API docs for NSX-T <https://code.vmware.com/apis/1163/nsx-t>`_

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    compute_managers_result = __salt__["nsxt_compute_manager.get"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        server=compute_manager_server,
    )
    if "error" in compute_managers_result:
        ret["result"] = False
        ret["comment"] = "Failed to get compute managers from NSX-T Manager : {}".format(
            compute_managers_result["error"]
        )
        return ret

    if compute_managers_result["result_count"] > 1:
        ret["result"] = False
        ret["comment"] = "Found multiple results for the provided compute manager in NSX-T"
        return ret
    else:
        existing_compute_manager = (
            compute_managers_result["results"][0]
            if compute_managers_result["result_count"] == 1
            else None
        )

    if __opts__["test"]:
        if existing_compute_manager is None:
            ret["result"] = None
            ret["comment"] = "Compute manager would be registered to NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "Compute manager registration would be updated"
        return ret

    if isinstance(credential, dict):
        thumbprint = credential.get("thumbprint", None)
    else:
        ret["result"] = False
        ret[
            "comment"
        ] = "Parameter credential must be of type dictionary. Please refer documentation"
        return ret

    if existing_compute_manager is None:
        log.info("Going to register new compute manager, as no existing compute manager found")
        result = __salt__["nsxt_compute_manager.register"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            compute_manager_server=compute_manager_server,
            credential=credential,
            server_origin_type=server_origin_type,
            display_name=display_name,
            description=description,
            set_as_oidc_provider=set_as_oidc_provider,
        )
        if "error" in result:
            ret["result"] = False
            ret["comment"] = "Failed to register compute manager with NSX-T Manager : {}".format(
                result["error"]
            )
            return ret
        else:
            ret[
                "comment"
            ] = "Compute manager {compute_manager_server} successfully registered with NSX-T".format(
                compute_manager_server=compute_manager_server
            )
            ret["changes"]["new"] = result
            return ret
    else:
        log.info("Compute manager already exists. Going to check for updates")
        is_update_required = _needs_update(
            existing_compute_manager=existing_compute_manager,
            thumbprint=thumbprint,
            server_origin_type=server_origin_type,
            display_name=display_name,
            description=description,
            set_as_oidc_provider=set_as_oidc_provider,
        )
        if is_update_required:
            result = __salt__["nsxt_compute_manager.update"](
                hostname=hostname,
                username=username,
                password=password,
                verify_ssl=verify_ssl,
                cert=cert,
                cert_common_name=cert_common_name,
                compute_manager_server=compute_manager_server,
                credential=credential,
                server_origin_type=server_origin_type,
                display_name=display_name,
                description=description,
                set_as_oidc_provider=set_as_oidc_provider,
                compute_manager_id=existing_compute_manager["id"],
                compute_manager_revision=existing_compute_manager["_revision"],
            )
            if "error" in result:
                ret["result"] = False
                ret[
                    "comment"
                ] = "Failed to update existing registration of compute manager with NSX-T Manager : {}".format(
                    result["error"]
                )
                return ret
            else:
                ret[
                    "comment"
                ] = "Compute manager {compute_manager_server} registration successfully updated with NSX-T".format(
                    compute_manager_server=compute_manager_server
                )
                ret["changes"]["old"] = existing_compute_manager
                ret["changes"]["new"] = result
                return ret
        else:
            ret["comment"] = (
                "Compute manager registration for {compute_manager_server} already exists. "
                "Nothing to update. Modifiable fields:[display_name, description, set_as_oidc_provider, "
                "thumbprint, origin_type]".format(compute_manager_server=compute_manager_server)
            )
            return ret


def absent(
    name,
    hostname,
    username,
    password,
    compute_manager_server=None,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """

    De-Registers compute manager in NSX-T Manager if present. Requires either compute_manager_server or
    compute_manager_id to be provided

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.absent hostname=nsxt-manager.local username=admin ...

    .. code-block:: yaml

        register_compute_manager:
          nsxt_compute_manager.absent:
            - name: Registration
              hostname: {{ pillar['nsxt_manager_hostname'] }}
              username: {{ pillar['nsxt_manager_username'] }}
              password: {{ pillar['nsxt_manager_password'] }}
              cert: {{ pillar['nsxt_manager_certificate'] }}
              verify_ssl: <False/True>
              compute_manager_server: <compute manager IP address or FQDN>

    name
        Name of the operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    compute_manager_server
        (Optional) Compute manager server IP address or FQDN

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    compute_managers_result = __salt__["nsxt_compute_manager.get"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        server=compute_manager_server,
    )
    compute_manager_to_delete = None
    if "error" in compute_managers_result:
        ret["result"] = False
        ret["comment"] = "Failed to get compute managers from NSX-T Manager : {}".format(
            compute_managers_result["error"]
        )
        return ret

    if compute_managers_result["result_count"] == 1:
        compute_manager_to_delete = compute_managers_result["results"][0]
    elif compute_managers_result["result_count"] > 1:
        ret["result"] = False
        ret["comment"] = "Found multiple results for the provided compute manager in NSX-T"
        return ret

    if __opts__["test"]:
        if compute_manager_to_delete is not None:
            ret["result"] = None
            ret["comment"] = "Compute manager would be removed/de-registered from NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "Compute manager registration does not exists"
        return ret

    if compute_manager_to_delete is None:
        ret["comment"] = "Compute manager server not present in NSX-T manager"
        return ret
    else:
        result = __salt__["nsxt_compute_manager.remove"](
            hostname=hostname,
            username=username,
            password=password,
            compute_manager_id=compute_manager_to_delete["id"],
            verify_ssl=verify_ssl,
            cert=cert,
        )
        if "error" in result:
            ret["result"] = False
            ret[
                "comment"
            ] = "Failed to remove registration of compute manager with NSX-T Manager : {}".format(
                result["error"]
            )
            return ret
        else:
            ret["comment"] = "Compute manager registration removed successfully from NSX-T manager"
            ret["changes"]["old"] = compute_manager_to_delete
            ret["changes"]["new"] = {}
            return ret



"""
Salt Module to manage VMware NSX-T configuration

"""
import json
import logging

from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)

__virtual_name__ = "nsxt_manager"

NSXT_MANAGER_BASE_URL = "https://{}/api/v1/configs/management"


def __virtual__():
    return __virtual_name__


def set_manager_config(
    hostname,
    publish_fqdns,
    revision,
    username,
    password,
    cert_common_name=None,
    verify_ssl=True,
    cert=None,
):
    """
    Set NSX-T Manager's config

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_manager.set_manager_config hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    publish_fqdns
        Boolean value to set as publish_fqdns

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """

    log.info("Configuration's current Revision: %s", revision)
    req_data = {"publish_fqdns": publish_fqdns, "_revision": revision}

    log.debug("Setting value of publish_fqdns to %s", publish_fqdns)
    url = NSXT_MANAGER_BASE_URL.format(hostname)

    return nsxt_request.call_api(
        method="put",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=req_data,
    )


def get_manager_config(
    hostname, username, password, cert_common_name=None, verify_ssl=True, cert=None
):
    """
    Get NSX-T Manager's config

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_manager.get_manager_config hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    publish_fqdns
        Boolean value to set as publish_fqdns

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    """

    url = NSXT_MANAGER_BASE_URL.format(hostname)

    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )



"""
Salt execution module for VMC org users.
Provides methods to get, add and remove user within a org.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)

__virtualname__ = "vmc_org_users"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def list_(
    hostname,
    refresh_key,
    org_id,
    expand_profile=False,
    include_group_ids_in_roles=False,
    page_limit=None,
    page_start=None,
    service_definition_id=None,
    verify_ssl=True,
    cert=None,
):
    """
    Returns list of users for the given organization.

    Please refer the `Get Org Users documentation <https://developer.vmware.com/apis/csp/csp-iam/latest/csp/gateway/am/api/v2/orgs/orgId/users/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_org_users.list hostname=console.cloud.vmware.com org_id=org_id  ...

    hostname
        The host name of CSP.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    org_id
        The ID of organization for which user list is retrieved.

    expand_profile
        (Optional) A boolean value to indicate if the response should be expanded with the user profile.

    include_group_ids_in_roles
        (Optional) A boolean value to indicate if the inherited roles in the response should indicate group information.

    page_limit: Integer
        (Optional) Maximum number of users to return in response.

    page_start: Integer
        (Optional) Specifies the index that the set of results will begin with.

    service_definition_id
        (Optional) Service definition ID used to filter users having access to the service.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Org users for the organization %s", org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}csp/gateway/am/api/v2/orgs/{org_id}/users".format(
        base_url=api_base_url, org_id=org_id
    )

    allowed_kwargs = ["serviceDefinitionId", "pageStart", "pageLimit"]
    if expand_profile == True:
        allowed_kwargs.append("expandProfile")
    if include_group_ids_in_roles == True:
        allowed_kwargs.append("includeGroupIdsInRoles")

    params = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_kwargs,
        expandProfile=expand_profile,
        includeGroupIdsInRoles=include_group_ids_in_roles,
        serviceDefinitionId=service_definition_id,
        pageStart=page_start,
        pageLimit=page_limit,
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=hostname,
        description="vmc_org_users.list",
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def search(
    hostname,
    refresh_key,
    org_id,
    user_search_term,
    expand_profile=False,
    include_group_ids_in_roles=False,
    verify_ssl=True,
    cert=None,
):
    """
    Search users in organization having username, firstName, lastName or email which "contains" search term.
    e.g. search for "test" will return test@vmware.com if test@vmware.com is part of the organization.
    Search results limited to first 20 results. Please refine the search term for accurate results.
    Organization members will receive basic user information.
    Organization owners and Service Owners (for organizations that have access to the service) will additionally
    receive role details of the users.

    Please refer the `Search Org Users documentation <https://developer.vmware.com/apis/csp/csp-iam/latest/csp/gateway/am/api/orgs/orgId/users/search/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_org_users.search hostname=console.cloud.vmware.com org_id=org_id  ...

    hostname
        The host name of CSP.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    org_id
        The ID of organization for which user list is retrieved.

    user_search_term
        The string to be searched within email or firstName or lastName or username.

    expand_profile
        (Optional) A boolean value to indicate if the response should be expanded with the user profile.

    include_group_ids_in_roles
        (Optional) A boolean value to indicate if the inherited roles in the response should indicate group information.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Searching for Org users in the  org %s", org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}csp/gateway/am/api/orgs/{org_id}/users/search".format(
        base_url=api_base_url, org_id=org_id
    )

    allowed_kwargs = ["userSearchTerm"]
    if expand_profile == True:
        allowed_kwargs.append("expandProfile")
    if include_group_ids_in_roles == True:
        allowed_kwargs.append("includeGroupIdsInRoles")

    params = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_kwargs,
        userSearchTerm=user_search_term,
        expandProfile=expand_profile,
        includeGroupIdsInRoles=include_group_ids_in_roles,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=hostname,
        description="vmc_org_users.search",
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def add(
    hostname,
    refresh_key,
    org_id,
    user_name,
    organization_roles,
    skip_notify=False,
    custom_roles=None,
    service_roles=None,
    skip_notify_registration=False,
    invited_by=None,
    custom_groups_ids=None,
    verify_ssl=True,
    cert=None,
):
    """
    Invite users to the given organization.

    Please refer the `Organization User Invitation documentation <https://developer.vmware.com/apis/csp/csp-iam/latest/csp/gateway/am/api/orgs/orgId/invitations/post/>`_ to get insight of functionality and input parameters

    CLI Example:

      .. code-block:: bash

          salt <minion id> vmc_org_users.invite hostname=console.cloud.vmware.com org_id=org_id  ...

    hostname
          The host name of CSP.

    refresh_key
          API Token of the user which is used to get the Access Token required for VMC operations.

    org_id
          The ID of organization to which the user should be invited.

    user_name
        username (e-mail) of user to be invited to the organization.

    organization_roles
        List of unique organization roles assigned to user.
        It can be specified in the below format

        .. code::

            "organization_roles": [
                        {
                            "name": "org_member"
                        }
                    ]
            where 'name' indicates the name of the organization role.

    skip_notify
        (Optional) Indicates if the notification have to be skipped.

    custom_roles
        (Optional) List of custom role assigned to a user.
        It can be specified in the below format

        .. code::

            "custom_roles": [
                        {
                            "name": "role_name"
                        }
                    ]
            where 'name' indicates the name of the custom role.

    service_roles
        (Optional) List of service roles to attach to a user.
        Below fields defines the properties of service roles.

        'serviceDefinitionLink': (String) (Optional)
            The link to the service definition.

        'serviceRoles': list
            It can be specified in the below format

            .. code::

                "serviceRoles": [
                        {
                            "name": "role_name"
                        }
                    ]
                where 'name' indicates the name of the service role.

    skip_notify_registration
        (Optional) Prevent sending mails to users that do not yet have a CSP profile.

    invited_by
        (Optional) Indicates the actual user who is inviting.

    custom_groups_ids
        (Optional) List of unique IDs of Custom Groups. Can have a maximum of 15 custom group IDs.

    verify_ssl
       (Optional) Option to enable/disable SSL verification. Enabled by default.
       If set to False, the certificate validation is skipped.

    cert
       (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
       The certificate can be retrieved from browser.

    Example values:

        .. code::

            {
                "skip_notify": true,
                "custom_roles": [
                    {
                        "name": "string"
                    }
                ],
                "service_roles": [
                    {
                        "serviceRoles": [
                            {
                                "name": "vmc-user:full"
                            },
                            {
                                "name": "nsx:cloud_admin"
                            },
                            {
                                "name": "nsx:cloud_auditor"
                            }
                        ],
                        "serviceDefinitionLink": "/csp/gateway/slc/api/definitions/paid/tcq4LTfyZ_-UPdPAJIi2LhnvxmE_"
                    }
                ],
                "skip_notify_registration": true,
                "organization_roles": [
                    {
                        "name": "org_member"
                    }
                ],
                "invited_by": "string",
                "user_name": "test@vmware.com",
                "custom_groups_ids": [
                    "string"
                ]
            }

        CLI Example:

          .. code-block:: bash

                salt <minion id> vmc_org_users.invite hostname=console.cloud.vmware.com org_id="1234" refresh_key="J05AftDxW" user_name="abc@example.com" organization_roles='[{"name": "org_member"},{"name": "developer"}]' service_roles='[{"serviceDefinitionLink": "/csp/gateway/slc/api/definitions/paid/tcq4LTfyZ_-UPdPAJIi2LhnvxmE_", "serviceRoles": [{"name": "vmc-user:full"}, {"name": "nsx:cloud_admin"}]}]' verify_ssl=false

    """

    log.info("Adding a user in the org %s", org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}csp/gateway/am/api/orgs/{org_id}/invitations".format(
        base_url=api_base_url, org_id=org_id
    )

    allowed_dict = {
        "usernames": [user_name],
        "skipNotify": skip_notify,
        "organizationRoles": organization_roles,
        "customRoles": custom_roles,
        "serviceRolesDtos": service_roles,
        "skipNotifyRegistration": skip_notify_registration,
        "invitedBy": invited_by,
        "customGroupsIds": custom_groups_ids,
    }

    req_data = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)

    request_data = vmc_request.create_payload_for_request(vmc_templates.add_org_users, req_data)

    result = vmc_request.call_api(
        method=vmc_constants.POST_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=hostname,
        description="vmc_org_users.add",
        responsebody_applicable=False,
        data=request_data,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    search_result = search(
        hostname,
        refresh_key=refresh_key,
        org_id=org_id,
        user_search_term=user_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )
    if search_result["results"]:
        log.info("Added %s in the org successfully %s", user_name, org_id)
        result.update(message="Added {} successfully".format(user_name))
    else:
        # check if user_name is in invitation list
        api_base_url = vmc_request.set_base_url(hostname)
        api_url = "{base_url}csp/gateway/am/api/orgs/{org_id}/invitations".format(
            base_url=api_base_url, org_id=org_id
        )

        allowed_kwargs = ["expand"]
        params = vmc_request._filter_kwargs(
            allowed_kwargs=allowed_kwargs,
            expand=True,
        )

        invitation_list = vmc_request.call_api(
            method=vmc_constants.GET_REQUEST_METHOD,
            url=api_url,
            refresh_key=refresh_key,
            authorization_host=hostname,
            description="vmc_org_users.invitation_list",
            params=params,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        for invitation in invitation_list:
            if user_name == invitation["username"] and invitation["status"] == "AVAILABLE":
                log.info("Invited %s successfully", user_name)
                result = {
                    "message": "Invited {} successfully".format(user_name),
                    "invitation": invitation,
                }
                break
        else:
            log.info("Failed to add %s", user_name)
            result = {vmc_constants.ERROR: "Failed to add {}".format(user_name)}
    return result


def remove(hostname, refresh_key, org_id, user_ids, notify_users=False, verify_ssl=True, cert=None):
    """
     Remove users from organization by user ids.
     User ids will be of the format : e.g. vmware.com:820e7ca5-4024-407e-8db4-f552d5d03403.
     Pay attention: in case of partial success the caller must read the response to see which users have not been removed successfully

    Please refer the `remove users from org documentation <https://developer.vmware.com/apis/csp/csp-iam/latest/csp/gateway/am/api/v2/orgs/orgId/users/delete/>`_ to get insight of functionality and input parameters

     CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_org_users.remove hostname=console.cloud.vmware.com org_id=org_id  ...

    hostname
        The host name of CSP.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    org_id
        The ID of organization from which the users should be removed.

    user_ids
        List of unique IDs of the users to be removed from the organization. Can have a maximum of 20 user IDs.

    notify_users
        (Optional) Indicates whether the users need to notify through email. Default value is true.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    For example:

        .. code::

            {
                "user_ids": [
                    "vmware.com:7109c7a3-fcef-452d-998e-3fab161bc0d6"
                ],
                "notify_users": False
            }

    """
    log.info("Removing a user in the org %s", org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}csp/gateway/am/api/v2/orgs/{org_id}/users".format(
        base_url=api_base_url, org_id=org_id
    )

    allowed_dict = {
        "ids": user_ids,
        "notifyUsers": notify_users,
    }
    req_data = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)
    request_data = vmc_request.create_payload_for_request(vmc_templates.remove_org_users, req_data)

    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=hostname,
        description="vmc_org_users.remove",
        data=request_data,
        verify_ssl=verify_ssl,
        cert=cert,
    )
